import importlib
import importlib.util

import wx

from . import art
from . import errors
from .persistence import get_json_data, save_json_data

import logging
log = logging.getLogger("action")
sync_log = logging.getLogger("sync")


global_action_ids = {
    "about": wx.ID_ABOUT,
    "quit": wx.ID_EXIT,
    "prefs": wx.ID_PREFERENCES,
}


def get_action_id(action_key):
    global global_action_ids

    try:
        id = global_action_ids[action_key]
    except KeyError:
        id = wx.NewId()
        global_action_ids[action_key] = id
    return id

found_modules = {}
ignore_modules = set()

def find_action_factory_in_module(mod_root, action_key):
    """Given a module name, search for the action using heuristics matching
    based on the action name.

    First, the module mod_root.action_key is searched for the class named
    action_key. If not found, the action_key is shortened in successive steps,
    breaking up the action name by underscores. If it is not found in any of
    those steps, it is searched for in mod_root.__init__.

    For example, if the module is "sawx.actions" and the action
    key is "file_save_as", the lookup order will be:

        sawx.actions.file_save_as.py
        sawx.actions.file_save.py
        sawx.actions.file.py
        sawx.actions.__init__.py
    """
    global found_modules, ignore_modules

    action_key_parts = action_key.split("_")
    modules = []
    while bool(action_key_parts):
        modules.append("." + "_".join(action_key_parts))
        action_key_parts.pop()
    modules.append("")
    for ext in modules:
        mod_name = f"{mod_root}{ext}"
        if mod_name in ignore_modules:
            log.debug(f"ignoring {mod_name}")
            continue
        try:
            mod = found_modules[mod_name]
            log.debug(f"found previously discovered {mod_name}")
        except KeyError:
            try:
                log.debug(f"importlib: find_spec({mod_name})")
                spec = importlib.util.find_spec(mod_name)
            except (AttributeError, ModuleNotFoundError):
                # non-existent or the name refers to a .py file instead of a
                # sub-module with a __init__.py
                log.debug(f"importlib: {mod_name} is not a module; ignoring")
                ignore_modules.add(mod_name)
                continue
            except Exception as e:
                log.error(f"syntax error in module {mod_name}: {e}")
                import traceback
                log.error(traceback.format_exc())
                ignore_modules.add(mod_name)
                continue
            else:
                if spec is None:
                    log.debug(f"non-existent module {mod_name}")
                    ignore_modules.add(mod_name)
                    continue
            try:
                mod = importlib.import_module(mod_name)
            except Exception as e:
                log.error(f"error loading module {mod_name}: {e}")
                import traceback
                log.error(traceback.format_exc())
                ignore_modules.add(mod_name)
                continue
            else:
                log.debug(f"found module {mod_name}")
                found_modules[mod_name] = mod

        try:
            factory = mod.action_factory
        except AttributeError:
            log.debug(f"no callable action_factory in {mod_name}")
        else:
            log.debug(f"found callable action_factory in {mod_name}")
            return factory
        log.debug(f"searching in {mod_name} for '{action_key}'")
        try:
            factory = getattr(mod, action_key)
        except AttributeError:
            log.debug(f"no callable {action_key} in {mod_name}")
        else:
            log.debug(f"found callable {action_key} in {mod_name}")
            return factory

def find_action_factory(module_search_order, action_key):
    """Search for the action given its key.

    The module search order is a list of module names. Within each of those
    module names, the action is searched for using some heuristics based on the
    action name. See `find_action_factory_in_module` for more details.

    If no action is found, KeyError is raised.
    """
    for mod in module_search_order:
        action_factory = find_action_factory_in_module(mod, action_key)
        if action_factory is not None:
            return action_factory
    else:
        raise KeyError(f"no action factory found for {action_key} in {module_search_order}")

class SawxAction:
    name = "base action"
    icon = None

    def __init__(self, editor, action_key):
        self.editor = editor
        self.popup_data = None  # set by MenuDescription if action is in a popup menu
        self.init_from_editor(action_key)

    def calc_name(self, action_key):
        return self.name

    def calc_icon_name(self, action_key):
        return self.icon or action_key

    def perform_as_menu_item(self, action_key):
        # NOTE: call this using a CallAfter
        log.debug(f"performing as menu item: {action_key}")
        try:
            self.perform(action_key)
        except RuntimeError as e:
            log.error(e)
            self.editor.frame.error(str(e))

    def perform_as_keystroke(self, action_key):
        log.debug(f"performing as keystroke: {action_key}")
        try:
            self.perform(action_key)
        except AttributeError as e:
            raise errors.ProcessKeystrokeNormally(e)
        except RuntimeError as e:
            log.error(e)
            self.editor.frame.error(str(e))

    def perform(self, action_key):
        raise AttributeError(f"no perform method defined for {self}")

    def calc_menu_sub_keys(self, action_key):
        """Return any dynamically created action_keys

        Actions that create multiple entries in the menubar/toolbar must create
        unique action_keys here."""
        raise AttributeError("this action has no dynamically created sub-actions")

    def append_to_menu(self, menu, id, action_key):
        name = self.calc_name(action_key)
        log.debug(f"append_to_menu: {action_key}: name={name}")
        menu.Append(id, name)

    def append_to_toolbar(self, tb, id, action_key):
        name = self.calc_name(action_key)
        log.debug(f"append_to_toolbar: {action_key}: name={name}")
        tb.AddTool(id, name, self.calc_bitmap(action_key), wx.NullBitmap, wx.ITEM_NORMAL, name, f"{action_key} id={id} '{name}'", None)

    def calc_bitmap(self, action_key):
        icon_name = self.calc_icon_name(action_key)
        return art.get_bitmap(icon_name, self.editor.tool_bitmap_size)

    def init_prefix(self, action_key):
        if "{" in action_key and action_key.endswith("}"):
            _, action_list_id = action_key[:-1].split("{", 1)
        else:
            action_list_id = ""
        log.debug(f"action_key={action_key}, list_id={action_list_id}")
        self.action_list_id = action_list_id
        self.prefix = action_key + "_"
        self.prefix_count = len(self.prefix)

    def init_from_editor(self, action_key):
        pass

    def calc_enabled(self, action_key):
        return True

    def sync_menu_item_from_editor(self, action_key, menu_item):
        state = self.calc_enabled(action_key)
        menu_item.Enable(state)

    def sync_tool_item_from_editor(self, action_key, toolbar_control, id):
        state = self.calc_enabled(action_key)
        toolbar_control.EnableTool(id, state)


class SawxSubAction(SawxAction):
    def init_from_editor(self, action_key):
        self.init_prefix(action_key)

    def calc_name(self, action_key):
        return self.action_list_id

    def calc_icon_name(self, action_key):
        return self.prefix


class SawxNameChangeAction(SawxAction):
    def sync_menu_item_from_editor(self, action_key, menu_item):
        SawxAction.sync_menu_item_from_editor(self, action_key, menu_item)
        name = self.calc_name(action_key)
        menu_item.SetText(name)


class SawxActionRadioMixin:
    def append_to_menu(self, menu, id, action_key):
        menu.AppendCheckItem(id, self.calc_name(action_key))

    def append_to_toolbar(self, tb, id, action_key):
        name = self.calc_name(action_key)
        tb.AddTool(id, name, self.calc_bitmap(action_key), wx.NullBitmap, wx.ITEM_RADIO, name, f"{action_key} id={id} '{name}'", None)


class SawxRadioAction(SawxActionRadioMixin, SawxAction):
    def calc_checked(self, action_key):
        """Return checked state of item

        If the item passed into this function is different than the item from
        the source, raise `RecreateDynamicMenuBar` error to force recalculation
        of the whole menubar.
        """
        raise NotImplementedError

    def sync_menu_item_from_editor(self, action_key, menu_item):
        if self.calc_enabled(action_key):
            state = self.calc_checked(action_key)
            sync_log.debug(f"{action_key}: checked={state}")
            menu_item.Enable(True)
            menu_item.Check(state)
        else:
            menu_item.Enable(False)

    def sync_tool_item_from_editor(self, action_key, toolbar_control, id):
        if self.calc_enabled(action_key):
            state = self.calc_checked(action_key)
            sync_log.debug(f"{action_key}: checked={state}")
            toolbar_control.EnableTool(id, True)
            toolbar_control.ToggleTool(id, state)
        else:
            toolbar_control.EnableTool(id, False)


class SawxActionListMixin:
    empty_list_name = "No Items"

    def init_from_editor(self, action_key):
        self.init_prefix(action_key)
        self.current_list = self.calc_list_items()

    def calc_list_items(self):
        return []

    def get_index(self, action_key):
        return int(action_key[self.prefix_count:])

    def get_item(self, action_key):
        return self.current_list[self.get_index(action_key)] if self.current_list else None

    def calc_name(self, action_key):
        if len(self.current_list) == 0:
            return self.empty_list_name
        return str(self.get_item(action_key))

    def calc_menu_sub_keys(self, action_key):
        if len(self.current_list) == 0:
            return [self.prefix + "empty"]
        return [f"{self.prefix}{i}" for i in range(len(self.current_list))]


class SawxListAction(SawxActionListMixin, SawxAction):
    def sync_menu_item_from_editor(self, action_key, menu_item):
        if self.current_list:
            if self.get_index(action_key) == 0:
                # Only need to check one action key because they all use the
                # same list of items
                new_list = self.calc_list_items()
                if new_list != self.current_list:
                    raise errors.RecreateDynamicMenuBar
            state = self.calc_enabled(action_key)
            menu_item.Enable(state)
        else:
            if self.calc_list_items():
                raise errors.RecreateDynamicMenuBar
            menu_item.Enable(False)

    def perform(self, action_key):
        raise NotImplementedError


persistent_dict = {}

class SawxPersistentDictAction(SawxListAction):
    json_name = "test_action"

    @classmethod
    def get_dict(cls):
        global persistent_dict

        if cls.json_name not in persistent_dict:
            cls.reload()
        return persistent_dict[cls.json_name]

    @classmethod
    def reload(cls):
        global persistent_dict

        data = get_json_data(cls.json_name, {})
        persistent_dict[cls.json_name] = data

    def calc_list_items(self):
        d = self.get_dict()
        return sorted(d.keys())

    @classmethod
    def add(cls, label, data):
        d = cls.get_dict()
        d[label] = data
        save_json_data(cls.json_name, d)


class SawxRadioListAction(SawxActionListMixin, SawxActionRadioMixin, SawxAction):

    def calc_checked_list_item(self, action_key, index, item):
        """Return checked state of item

        If the item passed into this function is different than the item from
        the source, raise `RecreateDynamicMenuBar` error to force recalculation
        of the whole menubar.
        """
        raise NotImplementedError

    def sync_menu_item_from_editor(self, action_key, menu_item):
        if self.current_list:
            state = self.calc_enabled(action_key)
            if state:
                index = self.get_index(action_key)
                item = self.current_list[index]
                checked = self.calc_checked_list_item(action_key, index, item)
                sync_log.debug(f"{action_key}: checked={state}, {item}")
            else:
                checked = False
            menu_item.Enable(state)
            menu_item.Check(checked)
        else:
            if self.calc_list_items():
                raise errors.RecreateDynamicMenuBar
            menu_item.Enable(False)

    def sync_tool_item_from_editor(self, action_key, toolbar_control, id):
        if self.current_list:
            state = self.calc_enabled(action_key)
            if state:
                index = self.get_index(action_key)
                item = self.current_list[index]
                checked = self.calc_checked_list_item(action_key, index, item)
                sync_log.debug(f"{action_key}: checked={state}, {item}")
            else:
                checked = False
            toolbar_control.EnableTool(id, state)
            toolbar_control.ToggleTool(id, checked)
        else:
            if self.calc_list_items():
                raise errors.RecreateDynamicMenuBar
            toolbar_control.EnableTool(id, False)

    def perform(self, action_key):
        raise NotImplementedError



import os
import re
import shlex

from .runtime import get_all_subclasses
# from .file_guess import FileMetadata

import logging
log = logging.getLogger(__name__)


class HistoryList(list):
    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        self.insert_index = 0
        self.save_point_index = 0

    def __str__(self):
        text = ""
        for i, s in enumerate(self.history_list()):
            if i == self.insert_index:
                text += "  ---> insert index\n"
            text += s + "\n"
        if self.insert_index == len(self):
            text += "  ---> insert index\n"
        return text

    def is_dirty(self):
        return self.insert_index != self.save_point_index

    def set_save_point(self):
        self.save_point_index = self.insert_index

    @property
    def can_undo(self):
        return self.insert_index > 0

    @property
    def can_redo(self):
        return self.insert_index < len(self)

    def history_list(self):
        h = [str(c) for c in self]
        return h

    def get_undo_command(self):
        if self.can_undo:
            return self[self.insert_index - 1]

    def get_redo_command(self):
        if self.can_redo:
            return self[self.insert_index]

    def add_command(self, command):
        self[self.insert_index:] = [command]
        self.insert_index += 1

    def prev_command(self):
        cmd = self.get_undo_command()
        if cmd is not None:
            self.insert_index -= 1
        return cmd

    def next_command(self):
        cmd = self.get_redo_command()
        if cmd is not None:
            self.insert_index += 1
        return cmd


class UndoStack(HistoryList):
    def __init__(self, *args, **kwargs):
        HistoryList.__init__(self, *args, **kwargs)
        self.batch = self

    def perform_setup(self, editor):
        pass

    def perform(self, cmd, editor, batch=None):
        if cmd is None:
            return UndoInfo(editor)
        if batch != self.batch:
            if batch is None:
                self.end_batch()
            else:
                if self.batch is not None:
                    self.end_batch()
                self.start_batch(batch)
        self.batch.perform_setup(editor)
        undo_info = UndoInfo(editor)
        cmd.perform(editor, undo_info)
        if undo_info.flags.changed_document:
            if undo_info.flags.success:
                self.add_command(cmd)
            cmd.last_flags = undo_info.flags
        return undo_info

    def undo(self, editor):
        cmd = self.get_undo_command()
        if cmd is None:
            return UndoInfo(editor)
        undo_info = cmd.undo(editor)
        if undo_info.flags.success:
            self.insert_index -= 1
        cmd.last_flags = undo_info.flags
        return undo_info

    def redo(self, editor):
        cmd = self.get_redo_command()
        if cmd is None:
            return UndoInfo(editor)
        undo_info = UndoInfo(editor)
        cmd.perform(editor, undo_info)
        if undo_info.flags.success:
            self.insert_index += 1
        cmd.last_flags = undo_info.flags
        return undo_info

    def start_batch(self, batch):
        self.batch = batch
        log.debug("Starting batch %s" % self.batch)

    def end_batch(self):
        log.debug("Ending batch %s" % self.batch)
        if self.batch != self:
            batch_command = self.batch.get_recordable_command()
            if batch_command is not None:
                self.batch = self
                self.add_command(batch_command)

    def add_command(self, command):
        self.batch.insert_at_index(command)

    def insert_at_index(self, command):
        last = self.get_undo_command()
        if last is not None and last.coalesce(command):
            return
        if command.is_recordable():
            self[self.insert_index:] = [command]
            self.insert_index += 1

    def pop_command(self):
        last = self.get_undo_command()
        if last is not None:
            self.insert_index -= 1
            self[self.insert_index:self.insert_index + 1] = []
        return last

    def serialize(self):
        s = Serializer()
        for c in self:
            s.add(c)
        return s

    def unserialize_text(self, text, manager):
        offset = manager.get_invariant_offset()
        s = TextDeserializer(text, offset)
        for cmd in s.iter_cmds(manager):
            yield cmd

    def find_most_recent(self, cmdcls):
        for cmd in reversed(self):
            if isinstance(cmd, cmdcls):
                return cmd
        return None


class DisplayFlags:
    def __init__(self, source_control=None, args=[]):
        # True if command successfully completes, must set to False on failure
        self.success = True

        # True if command made a change to the document and therefore should be recorded
        self.changed_document = True

        # List of errors encountered
        self.errors = []

        # Message displayed to the user
        self.message = ""

        # has any data values changed, forcing all views to be refreshed?
        self.byte_values_changed = False

        # has any data style changed, forcing all views to be refreshed?
        self.byte_style_changed = False

        # has anything in the data or metadata changed to require a rebuild
        # of the data model?
        self.data_model_changed = False

        # set to True if the all views of the data need to be refreshed
        self.refresh_needed = False

        # ensure the specified index is visible
        self.index_visible = None

        # ensure the specified index range is visible
        self.index_range = None

        # set to True if the index_range should be selected
        self.select_range = False

        # set caret index to position
        self.caret_index = None

        # keep any selection instead of erasing during a caret move
        self.keep_selection = None

        # set if document properties have changed, but not the actual data
        self.metadata_dirty = None

        # set if user interface needs to be updated (very heavyweight call!)
        self.rebuild_ui = None

        # the source control on which the event happened, if this is the
        # result of a user interface change
        self.source_control = source_control

        # if the source control is refreshed as a side-effect of some action,
        # set this flag so that the event manager can skip that control when
        # it refreshes the others
        self.skip_source_control_refresh = False

        # if the portion of the window looking at the data needs to be changed,
        # these will be the new upper left coordinates
        self.viewport_origin = None

        # list of viewers that have been refreshed during the caret_flags
        # processing so it won't be updated again
        self.refreshed_as_side_effect = set()

        # set if the user is selecting by entire rows
        self.selecting_rows = False

        # if not None, will contain the set carets to determine if any have
        # moved and need to be updated.
        self.old_carets = None

        # if True will add the old carets to the current caret to increase the
        # number of carets by one
        self.add_caret = False

        # if True will remove all carets except the current caret
        self.force_single_caret = False

        # move the caret(s) to the next edit position (usually column) using
        # the control as the basis for how much the index needs to be adjusted
        # to get to the next column.
        self.advance_caret_position_in_control = None

        # sync the carets in all other controls from the given control.
        self.sync_caret_from_control = None

        if args:
            for flags in args:
                self.add_flags(flags)

    def __str__(self):
        flags = []
        flags.append(f"source_control={self.source_control}")
        for name in dir(self):
            if name.startswith("_"):
                continue
            val = getattr(self, name)
            if val is None or not val or hasattr(val, "__call__"):
                continue
            flags.append("%s=%s" % (name, val))
        return ", ".join(flags)

    def add_flags(self, flags, cmd=None):
        if flags.message is not None:
            self.message += flags.message
        if flags.errors:
            if cmd is not None:
                self.errors.append("In %s:" % str(cmd))
            for e in flags.errors:
                self.errors.append("  %s" % e)
            self.errors.append("")
        if flags.byte_values_changed:
            self.byte_values_changed = True
        if flags.byte_style_changed:
            self.byte_style_changed = True
        if flags.refresh_needed:
            self.refresh_needed = True
        if flags.select_range:
            self.select_range = True
        if flags.metadata_dirty:
            self.metadata_dirty = True
        if flags.rebuild_ui:
            self.rebuild_ui = True

        # Expand the index range to include the new range specified in flags
        if flags.index_range is not None:
            if self.index_range is None:
                self.index_range = flags.index_range
            else:
                s1, s2 = self.index_range
                f1, f2 = flags.index_range
                if f1 < s1:
                    s1 = f1
                if f2 > s2:
                    s2 = f1
                self.index_range = (s1, s2)

        if flags.caret_index is not None:
            self.caret_index = flags.caret_index
        if flags.force_single_caret:
            self.force_single_caret = flags.force_single_caret
        if flags.keep_selection:
            self.keep_selection = flags.keep_selection
        if flags.source_control:
            self.source_control = flags.source_control
        if flags.advance_caret_position_in_control:
            self.advance_caret_position_in_control = flags.advance_caret_position_in_control
        if flags.sync_caret_from_control:
            self.sync_caret_from_control = flags.sync_caret_from_control


class UndoInfo(object):
    def __init__(self, editor):
        self.index = -1
        self.data = None
        self.flags = editor.calc_status_flags()

    def __str__(self):
        return "index=%d, flags=%s" % (self.index, str(dir(self.flags)))


class Command(object):
    short_name = None
    ui_name = "<unnamed command>"
    serialize_order = [
        ]

    def __init__(self):
        self.undo_info = None
        self.last_flags = None

    def __str__(self):
        return self.ui_name

    def get_serialized_name(self):
        if self.short_name is None:
            return self.__class__.__name__
        return self.short_name

    def can_coalesce(self, next_command):
        """Evaluate whether or not the next command can be merged into this
        command.

        The difference between this and coalesce below is coalesce will take
        care of the very basics of comparisons and only send on commands that
        are of the same class.
        """
        return False

    def coalesce_merge(self, next_command):
        """Merge the next command into self

        Takes the details of next_command and combines them into the current
        instance. This is very implementation dependent, but the key is that
        the merged command must be undoable to the state before the current
        command.
        """
        raise NotImplementedError

    def coalesce(self, next_command):
        """If the next command can be merged with this one, merge them.

        Checks if the next command can be merged, and if so will merge the
        details of the next command into self. The default implementation calls
        can_coalesce to check if it can be merged, and if so calls
        coalesce_merge to actually merge the commands.
        """
        if next_command.__class__ == self.__class__:
            if self.can_coalesce(next_command):
                self.coalesce_merge(next_command)

    def is_recordable(self):
        return True

    def perform_setup(self, document):
        pass

    def do_change(self, editor, undo_info):
        raise NotImplementedError

    def set_undo_flags(self, flags):
        pass

    def perform(self, editor, undo_info):
        old_data = self.do_change(editor, undo_info)
        undo_info.data = (old_data, )
        self.set_undo_flags(undo_info.flags)
        self.undo_info = undo_info

    def undo_change(self, editor, old_data):
        raise NotImplementedError

    def undo(self, editor):
        old_data, = self.undo_info.data
        self.undo_change(editor, old_data)
        return self.undo_info


class Batch(Command):
    """A batch is immutable once created, so there's no need to allow
    intermediate index points.
    """
    ui_name = "<batch>"

    def __init__(self):
        Command.__init__(self)
        self.commands = []

    def get_recordable_command(self):
        return self

    def perform(self, editor, undo_info):
        flags = DisplayFlags()
        for c in self.commands:
            undo = c.perform(editor)
            flags.add_flags(undo.flags)
        undo_info.flags = flags

    def undo(self, editor):
        flags = DisplayFlags()
        for c in reversed(self.commands):
            undo = c.undo(editor)
            flags.add_flags(undo.flags)
        undo = UndoInfo()
        undo.flags = flags
        return undo

    def insert_at_index(self, command):
        if command.is_recordable():
            self.commands.append(command)


class Overlay(Command):
    ui_name = "<overlay>"

    def __init__(self):
        Command.__init__(self)
        self.last_command = None

    def get_recordable_command(self):
        return self.last_command

    def perform_setup(self, editor):
        flags = DisplayFlags()
        last = self.last_command
        if last is not None:
            undo = last.undo(editor)
            flags.add_flags(undo.flags)
        undo = UndoInfo(editor)
        undo.flags = flags
        return undo

    def undo(self, editor):
        # Not used because the overlay is replaced by the last command.  See
        # get_recordable_command above
        pass

    def insert_at_index(self, command):
        self.last_command = command


def get_known_commands():
    s = get_all_subclasses(Command)
    c = {}
    for cls in s:
        if cls.short_name is not None:
            c[cls.short_name] = cls
    return c


# shlex quote routine modified from python 3 to allow [ and ] unquoted for lists
_find_unsafe = re.compile(r'[^\w@%+=:,./[\]-]').search


def quote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"
    if _find_unsafe(s) is None:
        return s
    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return "'" + s.replace("'", "'\"'\"'") + "'"


class UnknownCommandError(RuntimeError):
    pass


class Serializer(object):
    known_commands = None

    def __init__(self, magic_id, magic_version):
        self.magic_header = self.get_magic(magic_id, magic_version)
        self.serialized_commands = []

    @classmethod
    def get_magic(cls, template, version):
        """Return an identifier string used at the top of a file to indicate
        the version of the serialized file
        """
        return "# %s, v%d" % (identifier, version)

    def __str__(self):
        lines = [self.magic_header]
        for cmd in self.serialized_commands:
            lines.append(str(cmd))
        return "\n".join(lines)

    def add(self, cmd):
        sc = SerializedCommand(cmd)
        self.serialized_commands.append(sc)

    @classmethod
    def get_command(cls, short_name):
        if cls.known_commands is None:
            cls.known_commands = command.get_known_commands()
        try:
            return cls.known_commands[short_name]
        except KeyError:
            return UnknownCommandError(short_name)


class TextDeserializer(object):
    def __init__(self, text, magic_id, magic_version):
        lines = text.splitlines(True)
        self.header = lines.pop(0)
        self.lines = lines
        magic_template = Serializer.get_magic(magic_id, magic_version)
        if not self.header.startswith(magic_template):
            raise RuntimeError("Not a %s command file!" % magic_id)

    def iter_cmds(self, manager):
        build_multiline = ""
        for line in self.lines:
            if build_multiline:
                line = build_multiline + line
                build_multiline = ""
            else:
                if not line.strip() or line.startswith("#"):
                    continue
            try:
                text_args = shlex.split(line.strip())
            except ValueError as e:
                build_multiline = line
                continue
            cmd = self.unserialize_line(text_args, manager)
            yield cmd

    def unserialize_line(self, text_args, manager):
        short_name = text_args.pop(0)
        log.debug("unserialize: short_name=%s, args=%s" % (short_name, text_args))
        cmd_cls = Serializer.get_command(short_name)
        cmd_args = []
        for name, stype in cmd_cls.serialize_order:
            log.debug("  name=%s, type=%s" % (name, stype))
            converter = SerializedCommand.get_converter(stype)
            arg = converter.instance_from_args(text_args, manager, self)
            log.debug("  converter=%s: %s" % (converter.__class__.__name__, repr(arg)))
            cmd_args.append(arg)
        log.debug("COMMAND: %s(%s)" % (cmd_cls.__name__, ",".join([repr(a) for a in cmd_args])))
        cmd = cmd_cls(*cmd_args)
        return cmd


class ArgumentConverter(object):
    stype = None  # Default converter just uses strings

    def get_args(self, instance):
        """Return list of strings that can be used to reconstruct the instance
        """
        return str(instance),

    def instance_from_args(self, args, manager, deserializer):
        arg = args.pop(0)
        return arg


class FileMetadataConverter(ArgumentConverter):
    stype = "file_metadata"

    def get_args(self, instance):
        # Force forward slashes on windows so to prevent backslash escape chars
        return os.path.normpath(instance.uri).replace('\\', '/'), instance.mime

    def instance_from_args(self, args, manager, deserializer):
        uri = args.pop(0)
        mime = args.pop(0)
        return FileMetadata(uri=uri, mime=mime)


class TextConverter(ArgumentConverter):
    stype = "text"

    def get_args(self, instance):
        """Return list of strings that can be used to reconstruct the instance
        """
        return instance.encode("utf-8"),

    def instance_from_args(self, args, manager, deserializer):
        text = args.pop(0)
        return text.decode("utf-8")


class BoolConverter(ArgumentConverter):
    stype = "bool"

    def instance_from_args(self, args, manager, deserializer):
        text = args.pop(0)
        if text == "None":
            return None
        if text == "True" or text == "1":
            return True
        return False


class IntConverter(ArgumentConverter):
    stype = "int"

    def instance_from_args(self, args, manager, deserializer):
        text = args.pop(0)
        if text == "None":
            return None
        return int(text)


class FloatConverter(ArgumentConverter):
    stype = "float"

    def instance_from_args(self, args, manager, deserializer):
        text = args.pop(0)
        if text == "None":
            return None
        return float(text)


class PointConverter(ArgumentConverter):
    stype = "point"

    def get_args(self, instance):
        if instance is None:
            return None,
        return instance  # already a tuple

    def instance_from_args(self, args, manager, deserializer):
        lon = args.pop(0)
        if lon == "None":
            return None
        lat = args.pop(0)
        return (float(lon), float(lat))


class PointsConverter(ArgumentConverter):
    stype = "points"

    def get_args(self, instance):
        text = ",".join(["(%s,%s)" % (str(i[0]), str(i[1])) for i in instance])
        return "[%s]" % text,

    def instance_from_args(self, args, manager, deserializer):
        text = args.pop(0).lstrip("[").rstrip("]")
        if text:
            text = text.lstrip("(").rstrip(")")
            tuples = text.split("),(")
            points = []
            for t in tuples:
                lon, lat = t.split(",", 1)
                points.append((float(lon), float(lat)))
            return points
        return []


class RectConverter(ArgumentConverter):
    stype = "rect"

    def get_args(self, instance):
        (x1, y1), (x2, y2) = instance
        return x1, y1, x2, y2

    def instance_from_args(self, args, manager, deserializer):
        x1 = args.pop(0)
        y1 = args.pop(0)
        x2 = args.pop(0)
        y2 = args.pop(0)
        return ((x1, y1), (x2, y2))


class ListIntConverter(ArgumentConverter):
    stype = "list_int"

    def get_args(self, instance):
        text = ",".join([str(i) for i in instance])
        return "[%s]" % text,

    def instance_from_args(self, args, manager, deserializer):
        text = args.pop(0)
        if text.startswith("["):
            text = text[1:]
        if text.endswith("]"):
            text = text[:-1]
        if text:
            vals = text.split(",")
            return [int(i) for i in vals]
        return []


def get_converters():
    s = get_all_subclasses(ArgumentConverter)
    c = {}
    for cls in s:
        c[cls.stype] = cls()
    c[None] = ArgumentConverter()  # Include default converter
    return c


class SerializedCommand(object):
    converters = get_converters()

    def __init__(self, cmd):
        self.cmd_name = cmd.get_serialized_name()
        p = []
        for name, stype in [(n[0], n[1]) for n in cmd.serialize_order]:
            p.append((stype, getattr(cmd, name)))
        self.params = p

    def __str__(self):
        output = []
        for stype, value in self.params:
            try:
                c = self.converters[stype]
                values = c.get_args(value)
            except KeyError:
                values = [value]
            string_values = [quote(str(v)) for v in values]
            output.append(" ".join(string_values))

        text = " ".join(output)
        return "%s %s" % (self.cmd_name, text)

    @classmethod
    def get_converter(cls, stype):
        try:
            return cls.converters[stype]
        except KeyError:
            return cls.converters[None]



import sys
import time
import importlib
import inspect

import wx

from . import persistence
from .events import EventHandler
from .ui import fonts

import logging
log = logging.getLogger(__name__)


class cached_property:
    """Decorator for read-only properties evaluated only once until invalidated.

    It can be used to create a cached property like this::

        import random

        # the class containing the property must be a new-style class
        class MyClass(object):
            # create property whose value is cached for ten minutes
            @cached_property
            def randint(self):
                # will only be evaluated once unless invalidated
                return random.randint(0, 100)

    The value is cached  in the '_cache' attribute of the object instance that
    has the property getter method wrapped by this decorator. The '_cache'
    attribute value is a dictionary which has a key for every property of the
    object which is wrapped by this decorator. Each entry in the cache is
    created only when the property is accessed for the first time and is a
    two-element tuple with the last computed property value and the last time
    it was updated in seconds since the epoch.

    To expire a cached property value manually just do::

        del instance._cache[<property name>]

    """
    def __init__(self, func, doc=None):
        self.func = func
        self.__doc__ = doc or func.__doc__
        self.__name__ = func.__name__
        self.__module__ = func.__module__

    def __get__(self, inst, owner):
        try:
            value = inst._cache[self.__name__]
        except (KeyError, AttributeError):
            value = self.func(inst)
            try:
                cache = inst._cache
            except AttributeError:
                cache = inst._cache = {}
            cache[self.__name__] = value
        return value


class SawxPreferences:
    display_title = None

    display_order = [
    ]

    def __init__(self, module_path):
        self._module_path = module_path
        self.set_defaults()
        self.preferences_changed_event = EventHandler(self)

    def __eq__(self, other):
        if hasattr(other, "display_order") and self.display_order == other.display_order:
            for d in self.display_order:
                attrib_name = d[0]
                if getattr(self, attrib_name) != getattr(other, attrib_name):
                    break
            else:
                return True
        return False

    def set_defaults(self):
        self._image_caches = {}

    def calc_image_cache(self, cache_cls):
        try:
            c = self._image_caches[cache_cls]
        except KeyError:
            c = cache_cls(self)
            self._image_caches[cache_cls] = c
        return c

    def copy_from(self, other):
        for d in self.display_order:
            attrib_name = d[0]
            try:
                value = getattr(other, attrib_name)
            except AttributeError:
                try:
                    value = other[attrib_name]
                except KeyError:
                    continue
            try:
                setattr(self, attrib_name, value)
            except Exception:
                log.error(f"{self._module_path}: failed setting {attrib_name} to {value}")

    def clone(self):
        other = self.__class__(self._module_path)
        other.copy_from(self)
        return other

    def restore_user_settings(self):
        settings = persistence.get_json_data(self._module_path)
        log.debug(f"restore_user_overrides: user data = {settings}")
        if settings is not None:
            settings_dict = dict(settings)
            self.copy_from(settings_dict)

    def persist_user_settings(self):
        settings = [[d[0], getattr(self, d[0])] for d in self.display_order]
        log.debug(f"persist_user_settings: Saving {settings}")
        path = persistence.save_json_data(self._module_path, settings)
        log.debug(f"persist_user_settings: Saved to {path}")
        self.preferences_changed_event(True)



class SawxApplicationPreferences(SawxPreferences):
    ui_name = "General"

    display_order = [
        ("num_open_recent", "int", "Number of files in Open Recent menu"),
        ("confirm_before_close", "bool"),
    ]

    def set_defaults(self):
        SawxPreferences.set_defaults(self)
        self.num_open_recent = 15
        self.confirm_before_close = True


class SawxEditorPreferences(SawxPreferences):
    display_order = [
        ("text_font", "wx.Font"),
        ("text_color", "wx.Colour"),
        ("diff_text_color", "wx.Colour"),

        ("background_color", "wx.Colour"),
        ("highlight_background_color", "wx.Colour"),
        ("data_background_color", "wx.Colour"),
        ("match_background_color", "wx.Colour"),
        ("comment_background_color", "wx.Colour"),
        ("error_background_color", "wx.Colour"),
        ("empty_background_color", "wx.Colour"),

        ("unfocused_caret_color", "wx.Colour"),

        ("header_font", "wx.Font"),
        ("row_header_bg_color", "wx.Colour"),
        ("row_label_border_width", "int"),
        ("row_height_extra_padding", "int"),
        ("col_header_bg_color", "wx.Colour"),
        ("col_label_border_width", "int"),
        ("cell_padding_width", "int"),
    ]

    def set_defaults(self):
        SawxPreferences.set_defaults(self)
        self.text_font = fonts.str_to_font(fonts.default_font)
        self._background_color = wx.Colour(wx.WHITE)
        self.text_color = wx.Colour(wx.BLACK)
        self._empty_background_color = wx.Colour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE).Get(False))
        self.header_font = fonts.str_to_font(fonts.default_font + "bold")
        self.row_header_bg_color = wx.Colour(224, 224, 224)
        self.col_header_bg_color = wx.Colour(224, 224, 224)
        self.col_label_border_width = 3
        self.row_label_border_width = 3
        self.row_height_extra_padding = -3
        self.base_cell_width_in_chars = 2
        self.cell_padding_width = 2

        self.unfocused_caret_color = wx.Colour(128, 128, 128)
        self._highlight_background_color = wx.Colour(100, 200, 230)
        self._data_background_color = wx.Colour(224, 224, 224)
        self._match_background_color = wx.Colour(255, 255, 180)
        self._comment_background_color = wx.Colour(255, 180, 200)
        self.error_background_color = wx.Colour(255, 128, 128)
        self.diff_text_color = wx.Colour(255, 0, 0)

        self.caret_pen = wx.Pen(self.unfocused_caret_color, 1, wx.SOLID)

        self.selected_brush = wx.Brush(self.highlight_background_color, wx.SOLID)
        self.normal_brush = wx.Brush(self.background_color, wx.SOLID)
        self.data_brush = wx.Brush(self.data_background_color, wx.SOLID)
        self.match_brush = wx.Brush(self.match_background_color, wx.SOLID)
        self.comment_brush = wx.Brush(self.comment_background_color, wx.SOLID)
        self.empty_brush = wx.Brush(self.empty_background_color, wx.SOLID)

    @property
    def text_font(self):
        return self._text_font

    @text_font.setter
    def text_font(self, value):
        self._text_font = value
        dc = wx.MemoryDC()
        dc.SetFont(self._text_font)
        self.text_font_char_width = dc.GetCharWidth()
        self.text_font_char_height = dc.GetCharHeight()

    @property
    def background_color(self):
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        self._background_color = value
        self.normal_brush = wx.Brush(self.background_color, wx.SOLID)

    @property
    def highlight_background_color(self):
        return self._highlight_background_color

    @highlight_background_color.setter
    def highlight_background_color(self, value):
        self._highlight_background_color = value
        self.selected_brush = wx.Brush(self.highlight_background_color, wx.SOLID)

    @property
    def data_background_color(self):
        return self._data_background_color

    @data_background_color.setter
    def data_background_color(self, value):
        self._data_background_color = value
        self.data_brush = wx.Brush(self.data_background_color, wx.SOLID)

    @property
    def match_background_color(self):
        return self._match_background_color

    @match_background_color.setter
    def match_background_color(self, value):
        self._match_background_color = value
        self.match_brush = wx.Brush(self.match_background_color, wx.SOLID)

    @property
    def comment_background_color(self):
        return self._comment_background_color

    @comment_background_color.setter
    def comment_background_color(self, value):
        self._comment_background_color = value
        self.comment_brush = wx.Brush(self.comment_background_color, wx.SOLID)

    @property
    def empty_background_color(self):
        return self._empty_background_color

    @empty_background_color.setter
    def empty_background_color(self, value):
        self._empty_background_color = value
        self.empty_brush = wx.Brush(self.empty_background_color, wx.SOLID)

    def calc_cell_size_in_pixels(self, chars_per_cell):
        width = self.cell_padding_width * 2 + self.text_font_char_width * chars_per_cell
        height = self.row_height_extra_padding + self.text_font_char_height
        return width, height

    def calc_text_width(self, text):
        return self.text_font_char_width * len(text)


def find_application_preferences(prefs_module):
    mod = importlib.import_module(prefs_module)
    prefs_cls = SawxApplicationPreferences
    for name, obj in inspect.getmembers(mod):
        if inspect.isclass(obj) and SawxApplicationPreferences in obj.__mro__[1:]:
            prefs_cls = obj
            break
    return prefs_cls(prefs_module)


def find_editor_preferences(prefs_module):
    mod = importlib.import_module(prefs_module)
    prefs_cls = SawxEditorPreferences
    for name, obj in inspect.getmembers(mod):
        if inspect.isclass(obj) and SawxEditorPreferences in obj.__mro__[1:]:
            prefs_cls = obj
            break
    return prefs_cls(prefs_module)



import time

import wx
import wx.lib.inspection

from ..frame import SawxFrame
from ..action import SawxAction, SawxNameChangeAction, SawxListAction, SawxRadioAction
from ..persistence import iter_templates
from ..ui.dialogs import prompt_for_dec, get_file_dialog_wildcard
from ..ui.prefs_dialog import PreferencesDialog
from ..ui.error_logger import show_logging_frame
from .. import errors

import logging
log = logging.getLogger(__name__)


class new_window(SawxAction):
    name = "New Window"

    def perform(self, event=None):
        frame = SawxFrame(None, wx.GetApp().app_blank_uri)
        frame.Show()


class new_blank_file(SawxAction):
    prefix = "new_blank_file"

    def calc_name(self, action_key):
        return "Blank File"

    def perform(self, event=None):
        frame = self.editor.frame
        val = prompt_for_dec(frame, 'Enter file size in bytes', 'New Blank File', 256)
        if val > 256*256*16:
            if not frame.confirm(f"{val} bytes seems large. Are you sure you want to open a file this big?", "Confirm Large File"):
                val = 0
        if val is not None and val > 0:
            uri = "blank://%d" % val
            frame.load_file(uri)

class new_file_from_template(SawxListAction):
    prefix = "new_file_from_template_"

    canonical_list = None

    @classmethod
    def calc_list_items(cls):
        if cls.canonical_list is None:
            cls.canonical_list = cls.calc_canonical_list()
        return cls.canonical_list

    @classmethod
    def calc_canonical_list(cls):
        items = []
        for template in iter_templates("new file"):
            items.append(template)
        items.sort()
        return items

    def perform(self, action_key):
        frame = self.editor.frame
        item = self.get_item(action_key)
        frame.load_file(item["uri"])

class open_file(SawxAction):
    icon = "file"

    def calc_name(self, action_key):
        return "Open"

    def perform(self, action_key):
        frame = self.editor.frame
        path = frame.prompt_local_file_dialog()
        if path is not None:
            frame.load_file(path, self.editor)

class save_file(SawxAction):
    icon = "save"
    ext_list = [("All Documents", ".*")]

    def calc_name(self, action_key):
        return "Save"

    def calc_enabled(self, action_key):
        return self.editor.is_dirty

    def perform(self, action_key):
        e = self.editor
        try:
            e.save_to_uri()
        except errors.ReadOnlyFilesystemError:
            self.perform_prompt(action_key)

    def perform_prompt(self, action_key):
        e = self.editor
        path = e.frame.prompt_local_file_dialog("Save As", save=True, default_filename=e.document.name, wildcard=get_file_dialog_wildcard(self.ext_list))
        if path is not None:
            e.save_to_uri(path)

class save_as(save_file):
    def calc_name(self, action_key):
        return "Save As"

    def calc_enabled(self, action_key):
        return True

    perform = save_file.perform_prompt

class save_as_image(save_as):
    ext_list = [("PNG Images", ".png"), ("JPEG Images", ".jpg")]

    def calc_name(self, action_key):
        return "Save As Image"

    def perform(self, action_key):
        e = self.editor
        try:
            raw_data = e.get_numpy_image_before_prompt()
        except RuntimeError:
            # cancelled
            return
        path = e.frame.prompt_local_file_dialog("Save As Image", save=True, default_filename=e.document.root_name, wildcard=get_file_dialog_wildcard(self.ext_list))
        if path is not None:
            e.save_as_image(path, raw_data)

class quit(SawxAction):
    def calc_name(self, action_key):
        return "Quit"

    def perform(self, action_key):
        wx.GetApp().quit()

class undo(SawxNameChangeAction):
    def calc_name(self, action_key):
        command = self.editor.document.undo_stack.get_undo_command()
        if command is None:
            label = "Undo"
        else:
            text = str(command).replace("&", "&&")
            label = "Undo: " + text
        return label

    def calc_enabled(self, action_key):
        return self.editor.document.undo_stack.can_undo

    def perform(self, action_key):
        self.editor.undo()

class redo(SawxNameChangeAction):
    def calc_name(self, action_key):
        command = self.editor.document.undo_stack.get_redo_command()
        if command is None:
            label = "Redo"
        else:
            text = str(command).replace("&", "&&")
            label = "Redo: " + text
        return label

    def calc_enabled(self, action_key):
        return self.editor.document.undo_stack.can_redo

    def perform(self, action_key):
        self.editor.redo()

class cut(SawxAction):
    def calc_name(self, action_key):
        return "Cut"

    def calc_enabled(self, action_key):
        return self.editor.can_cut

    def perform(self, action_key):
        self.editor.copy_selection_to_clipboard()
        self.editor.delete_selection()

class copy(cut):
    def calc_name(self, action_key):
        return "Copy"

    def perform(self, action_key):
        self.editor.copy_selection_to_clipboard()

class paste(SawxAction):
    def calc_name(self, action_key):
        return "Paste"

    def calc_enabled(self, action_key):
        return self.editor.can_paste

    def perform(self, action_key):
        self.editor.paste_clipboard()

class delete_selection(SawxAction):
    def calc_name(self, action_key):
        return "Delete Selection"

    def calc_enabled(self, action_key):
        return self.editor.can_cut

    def perform_as_keystroke(self, action_key):
        if self.calc_enabled(action_key):
            self.perform(action_key)
        else:
            raise errors.ProcessKeystrokeNormally("No selection")

    def perform(self, action_key):
        self.editor.delete_selection()

class select_all(SawxAction):
    def calc_name(self, action_key):
        return "Select All"

    def calc_enabled(self, action_key):
        return True

    def perform(self, action_key):
        self.editor.select_all()

class select_none(SawxAction):
    def calc_name(self, action_key):
        return "Select None"

    def calc_enabled(self, action_key):
        return True

    def perform(self, action_key):
        self.editor.select_none()

class select_invert(SawxAction):
    def calc_name(self, action_key):
        return "Invert Selection"

    def calc_enabled(self, action_key):
        return True

    def perform(self, action_key):
        self.editor.select_invert()

class prefs(SawxAction):
    def calc_name(self, action_key):
        return "Preferences"

    def perform(self, action_key):
        wx.GetApp().show_preferences_dialog(self.editor.frame, self.editor.ui_name)

class about(SawxAction):
    def calc_name(self, action_key):
        return "About"

    def perform(self, action_key):
        wx.GetApp().show_about_dialog()

class prev_line(SawxAction):
    def calc_name(self, action_key):
        return "Previous Line"

    def perform(self, action_key):
        print("Up!")

class next_line(SawxAction):
    def calc_name(self, action_key):
        return "Next Line"

    def perform(self, action_key):
        print("Down!")

class prev_char(SawxAction):
    def calc_name(self, action_key):
        return "Previous Char"

    def perform(self, action_key):
        print("Left!")

class next_char(SawxAction):
    def calc_name(self, action_key):
        return "Next Char"

    def perform(self, action_key):
        print("Right!")

class show_toolbar(SawxRadioAction):
    name = "Show Toolbar"

    def calc_checked(self, action_key):
        return self.editor.show_toolbar

    def perform(self, action_key):
        self.editor.show_toolbar = not self.editor.show_toolbar
        wx.CallAfter(self.editor.frame.sync_active_tab)

class raise_exception(SawxAction):
    def calc_name(self, action_key):
        return "Raise Exception"

    def perform(self, action_key):
        val = int("this will raise a ValueError")

class test_progress(SawxAction):
    def calc_name(self, action_key):
        return "Test Progress Dialog"

    def perform(self, action_key):
        progress_log = logging.getLogger("progress")
        progress_log.info("START=First Test")
        try:
            progress_log.info("TITLE=Starting timer")
            for i in range(1000):
                print(i)
                progress_log.info("Trying %d" % i)
                for j in range(10):
                    time.sleep(.001)
                    wx.Yield()
                if i > 4:
                    progress_log.info("TIME_DELTA=Finished trying %d" % i)
                progress_log.info("PULSE")
                wx.Yield()

        except errors.ProgressCancelError as e:
            error = str(e)
        finally:
            progress_log.info("END")

class show_debug_log(SawxAction):
    def calc_name(self, action_key):
        return "View Error Log"

    def perform(self, action_key):
        show_logging_frame()

class widget_inspector(SawxAction):
    def calc_name(self, action_key):
        return "View Widget Inspector"

    def perform(self, action_key):
        wx.lib.inspection.InspectionTool().Show()

class show_focus(SawxAction):
    def calc_name(self, action_key):
        return "Show Control With Focus"

    def perform(self, action_key):
        focused = wx.Window.FindFocus()
        info = f"{focused.__class__.__name__}, {hex(id(focused))}, {focused.GetName()}, {focused.GetParent().GetName()}"
        print(info)
        self.editor.frame.status_message(info)



import logging
log = logging.getLogger(__name__)


class SawxError(RuntimeError):
    pass

class RecreateDynamicMenuBar(SawxError):
    pass

class ProcessKeystrokeNormally(SawxError):
    pass

class EditorNotFound(SawxError):
    pass

class UnsupportedFileType(SawxError):
    pass

class ProgressCancelError(SawxError):
    pass

class DocumentError(SawxError):
    pass

class MissingDocumentationError(SawxError):
    pass

class ClipboardError(SawxError):
    pass

class ReadOnlyFilesystemError(SawxError):
    pass



import sys

import numpy as np
import wx
import wx.lib.scrolledpanel as scrolled
import wx.lib.buttons as buttons

import logging
log = logging.getLogger(__name__)


class TileCategory(object):
    def __init__(self, name):
        self.name = name

    def get_height(self, parent):
        w, h = parent.GetTextExtent(self.name)
        return h + 5

    def draw(self, font, dc, rect):
        dc.DrawLabel(self.name, rect, wx.ALIGN_LEFT | wx.ALIGN_CENTER_VERTICAL)


class Tile(object):
    def __init__(self, tile_num, keystroke):
        self.tile_num = tile_num
        self.keystroke = keystroke

    def get_bytes(self):
        return (self.tile_num, )  # has to be tuple so it's hashable

    def get_height(self, parent):
        if parent.editor:  # Might get called before an editor is set
            return parent.editor.machine.antic_font.get_height(parent.zoom)
        return 10

    def draw(self, parent, dc, rect):
        if parent.editor:  # Might get called before an editor is set
            bmp = parent.editor.machine.antic_font.get_image(self.tile_num, parent.zoom)
            dc.DrawBitmap(bmp, rect.x+10, rect.y)


class TileListBox(wx.VListBox):
    # This method must be overridden.  When called it should draw the
    # n'th item on the dc within the rect.  How it is drawn, and what
    # is drawn is entirely up to you.
    def OnDrawItem(self, dc, rect, n):
        if self.GetSelection() == n:
            c = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT)
        else:
            c = self.GetForegroundColour()
        dc.SetFont(self.GetFont())
        dc.SetTextForeground(c)
        item = self.GetParent().items[n]
        item.draw(self.GetParent(), dc, rect)

    # This method must be overridden.  It should return the height
    # required to draw the n'th item.
    def OnMeasureItem(self, n):
        item = self.GetParent().items[n]
        h = item.get_height(self.GetParent())
        return h


class TileListControl(wx.Panel):
    """
    View for displaying categories of tiles to paint the map
    """

    def __init__(self, parent, task, command=None, **kwargs):
        wx.Panel.__init__(self, parent, -1, **kwargs)
        self.task = task
        self.editor = None
        self.command_cls = command

        self.tile_list = TileListBox(self, -1, style=wx.VSCROLL)
        self.tile_list.Bind(wx.EVT_LEFT_DOWN, self.on_tile_clicked)
        self.tile_list.Bind(wx.EVT_LISTBOX, self.on_tile)
        self.cat = wx.Choice(self, -1)
        self.cat.Bind(wx.EVT_CHOICE, self.on_category)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.cat, 0, wx.EXPAND, 0)
        sizer.Add(self.tile_list, 1, wx.EXPAND, 0)
        self.SetSizer(sizer)
        self.Fit()

        self.pattern_to_index = {}
        self.parse_tile_map([("test", np.arange(0,10, dtype=np.uint8))])
        self.current_tile = -1
        self.setup_tiles()
        self.zoom = 2

    def recalc_view(self):
        editor = self.task.active_editor
        if editor is not None:
            self.editor = editor
            self.parse_tile_map(editor.antic_tile_map)
            self.setup_tiles()

    def parse_tile_map(self, tile_map):
        self.tile_map = tile_map
        self.categories = []
        self.items = []
        for label, tiles in tile_map:
            item = TileCategory(label)
            self.categories.append((item, len(self.items)))
            self.items.append(item)
            for i in np.arange(np.alen(tiles)):
                item = Tile(tiles[i:i+1], "")
                self.items.append(item)
                self.pattern_to_index[item.get_bytes] = i

    def setup_tiles(self):
        cats = [c[0].name for c in self.categories]
        self.cat.Set(cats)
        self.cat.SetSelection(0)
        self.tile_list.SetItemCount(len(self.items))
        self.tile_list.SetSelection(self.current_tile)

    def on_category(self, event):
        cat = event.GetSelection()
        item, index = self.categories[cat]
        self.tile_list.ScrollToRow(index)

    def on_tile_clicked(self, event):
        p = event.GetPosition()
        index = self.tile_list.HitTest(p)
        e = self.editor
        if e is not None:
            self.current_tile = index
            wx.CallAfter(e.set_current_draw_pattern, self.items[index].get_bytes(), self)

    def show_pattern(self, pattern):
        index = self.index_from_pattern.get(pattern, None)
        if index is not None:
            self.tile_list.SetSelection(index)
            self.tile_list.ScrollToRow(index)

    def clear_tile_selection(self):
        self.current_tile = -1
        self.tile_list.SetSelection(-1)

    def on_tile(self, event):
        index = event.GetInt()
        self.change_tile(self.items[index])

    def change_tile(self, tile):
        e = self.editor
        if e is None:
            return
        cmd_cls = self.command_cls
        if cmd_cls is None:
            return
        if e.can_copy:
            index = e.anchor_start_index
        else:
            index = e.caret_index
        value = tile.get_bytes()
        cmd = cmd_cls(e.segment, index, index+len(value), value, True)
        e.process_command(cmd)


class TileButton(buttons.GenBitmapToggleButton):
    labelDelta = 0
    label_border = 4
    bgClr = wx.Colour(255, 255, 255)
    faceDnClr = wx.Colour(100, 200, 230)

    def DoGetBestSize(self):
        """
        Overridden base class virtual.  Determines the best size of the
        button based on the label and bezel size.
        """
        w, h, useMin = self._GetLabelSize()
        width = w + self.label_border + 2 * self.bezelWidth + 4 * int(self.useFocusInd)
        height = h + self.label_border + 2 * self.bezelWidth + 4 * int(self.useFocusInd)
        return (width, height)

    def InitColours(self):
        pass

    def DrawBezel(self, dc, x1, y1, x2, y2):
        pass

    def GetBackgroundBrush(self, dc):
        """
        Returns the current :class:`wx.Brush` to be used to draw the button background.

        :param wx.DC `dc`: the device context used to draw the button background.
        """
        if self.up:
            col = self.bgColor
        else:
            col = self.faceDnClr
        brush = wx.Brush(col)
        return brush

    @classmethod
    def set_colors(cls, machine):
        cls.bgColor = wx.SystemSettings.GetColour(wx.SYS_COLOUR_BACKGROUND)
        cls.faceDnClr = wx.Colour(*machine.highlight_color)



import wx
import wx.adv
import wx.aui as aui
# import wx.lib.agw.aui as aui

from . import menubar
from . import toolbar
from . import statusbar
from . import keybindings
from . import errors
from . import editor as editor_module
from . import loader
from . import clipboard
from .ui import dialogs
from .document import identify_document
from .filesystem import fsopen as open
from . import filesystem


import logging
log = logging.getLogger(__name__)
sync_log = logging.getLogger("sync")
progress_log = logging.getLogger("progress")


class SawxFrame(wx.Frame):
    def __init__(self, editor, uri=None):
        wx.Frame.__init__(self, None , -1, uri or editor.title, size=wx.GetApp().last_window_size)

        self.create_icon()

        self.raw_menubar = wx.MenuBar()
        self.Bind(wx.EVT_MENU, self.on_menu)

        self.raw_toolbar = self.CreateToolBar(wx.TB_HORIZONTAL | wx.NO_BORDER | wx.TB_FLAT)

        self.raw_statusbar = statusbar.RawStatusBar(self)

        self.Bind(wx.EVT_SIZE, self.on_size)
        self.Bind(wx.EVT_CLOSE, self.on_close)
        self.Bind(wx.EVT_ACTIVATE, self.on_activate)
        self.Bind(wx.EVT_CHAR_HOOK, self.on_char_hook)

        if wx.Platform == "__WXMAC__":
            self.Bind(wx.EVT_MENU_OPEN, self.on_menu_open_mac)
        else:
            # Mac needs deferred SetMenuBar because that's where it moves the About
            # and Preferences to the Application menu. Other platforms set it
            # here and forget about it.
            self.SetMenuBar(self.raw_menubar)

            if wx.Platform == "__WXMSW__":
                self.Bind(wx.EVT_MENU_OPEN, self.on_menu_open_win)
            else:
                # EVT_MENU_OPEN fires every time the mouse moves to a different main
                # menu item, so this reduces it to only once: when the mouse
                # enters the menubar control itself. This doesn't seem to work on
                # other platforms, as they have native menubars which don't seem
                # to operate like regular wx.Window classes.
                self.raw_menubar.Bind(wx.EVT_ENTER_WINDOW, self.on_menu_open_linux)

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.notebook = self.create_notebook()
        sizer.Add(self.notebook, 1, wx.GROW)
        self.SetSizer(sizer)

        self.active_editor = None
        self.active_editor_can_paste = False
        self.pending_editor_close = None
        if uri is not None:
            self.load_file(uri)
        else:
            self.add_editor(editor)

    @property
    def editors(self):
        for index in range(self.notebook.GetPageCount()):
            control = self.notebook.GetPage(index)
            yield control.editor

    @property
    def is_dirty(self):
        state = False
        for e in self.editors:
            state |= e.is_dirty
        return state

    def create_notebook(self):
        notebook = aui.AuiNotebook(self, -1)

        # Use flat style for tabs
        art = aui.AuiSimpleTabArt()
        notebook.SetArtProvider(art)

        self.Bind(aui.EVT_AUINOTEBOOK_PAGE_CHANGED, self.on_page_changed)
        self.Bind(aui.EVT_AUINOTEBOOK_PAGE_CLOSE, self.on_page_closing)
        self.Bind(aui.EVT_AUINOTEBOOK_PAGE_CLOSED, self.on_page_closed)
        return notebook

    def create_icon(self):
        data = open(wx.GetApp().app_icon, 'rb')
        image = wx.Image(data)
        icon = wx.Icon()
        try:
            icon.CopyFromBitmap(wx.Bitmap(image))
            self.SetIcon(icon)
        except:
            log.error("AboutDialog: bad icon file: %s" % self.about_image)

    def create_menubar(self):
        log.debug(f"create_menubar: active editor={self.active_editor}")
        self.menubar = menubar.MenubarDescription(self, self.active_editor)

    def sync_menubar(self):
        try:
            self.menubar.sync_with_editor(self.raw_menubar)
        except errors.RecreateDynamicMenuBar:
            self.create_menubar()
            self.menubar.sync_with_editor(self.raw_menubar)
        if wx.Platform == "__WXMAC__":
            self.SetMenuBar(self.raw_menubar)

    def create_toolbar(self):
        log.debug(f"create_toolbar: active editor={self.active_editor}")
        self.toolbar = toolbar.ToolbarDescription(self, self.active_editor)
        self.raw_toolbar.Realize()

    def rebuild_toolbar(self):
        self.create_toolbar()
        self.toolbar.sync_with_editor(self.raw_toolbar)

    def sync_toolbar(self):
        try:
            self.toolbar.sync_with_editor(self.raw_toolbar)
        except errors.RecreateDynamicMenuBar:
            self.create_toolbar()
            self.toolbar.sync_with_editor(self.raw_toolbar)
        self.raw_toolbar.Show(self.active_editor.show_toolbar)

    def create_statusbar(self):
        log.debug(f"create_statusbar: active editor={self.active_editor}")
        self.statusbar = statusbar.StatusbarDescription(self, self.active_editor)
        self.SetStatusBar(self.raw_statusbar)

    def sync_statusbar(self):
        try:
            self.statusbar.sync_with_editor(self.raw_statusbar)
        except errors.RecreateDynamicMenuBar:
            self.create_statusbar()
            self.statusbar.sync_with_editor(self.raw_statusbar)

    def create_keybindings(self):
        log.debug(f"create_keybindings: active editor={self.active_editor}")
        # current_id_map = self.menubar.valid_id_map
        current_id_map = None
        self.keybindings = keybindings.KeyBindingDescription(self.active_editor, current_id_map)

    def set_title(self):
        app = wx.GetApp()
        title = f"{self.active_editor.title} - {app.app_name}"
        self.SetTitle(title)

    def sync_name(self):
        self.set_title()
        index = self.notebook.GetSelection()
        editor = self.find_editor_from_index(index)
        sync_log.debug(f"title={self.GetTitle()}, tab_name={editor.tab_name}")
        self.notebook.SetPageText(index, editor.tab_name)

    def sync_can_paste(self):
        """Check if the active editor can paste anything in the clipboard
        """
        self.active_editor_can_paste = clipboard.can_paste(self.active_editor.supported_clipboard_data)

    def sync_active_tab(self):
        # This function is called by a timer, so it's possible the frame has
        # been destroyed before the timer calls this. Checking for the truth
        # value of the wx object is supposed to fail if the C++ part has been
        # destroyed. We'll see how well it works. Otherwise, it will take a
        # try/except on RuntimeError
        log.debug("sync_active_tab: sync start")
        if self.notebook and self.notebook.GetPageCount() > 0:
            log.debug("sync_active_tab: sync_name")
            self.sync_name()
            log.debug("sync_active_tab: sync_toolbar")
            self.sync_toolbar()
            log.debug("sync_active_tab: sync_can_paste")
            self.sync_can_paste()
        log.debug("sync_active_tab: sync done")

    def add_editor(self, editor, args=None):
        editor.frame = self
        control = editor.create_control(self.notebook)
        editor.control = control
        control.editor = editor
        editor.create_event_bindings()
        editor.create_layout()
        if self.active_editor is not None and self.active_editor.is_transient:
            self.close_editor(self.active_editor)
        self.notebook.AddPage(control, editor.tab_name)
        self.make_active(editor)
        editor.show(args)

    def add_document(self, document, current_editor=None, args=None):
        if current_editor.__class__ == type:
            editor_cls = current_editor
        elif current_editor is not None and current_editor.can_edit_document(document):
            editor_cls = current_editor.__class__
        else:
            editor_cls = editor_module.find_editor_class_for_document(document)
        new_editor = editor_cls(document)
        log.debug(f"load_file: Created editor {new_editor}")
        # have to add before load so the control exists
        self.add_editor(new_editor, args)
        return new_editor

    def close_editor(self, editor, remove=True, quit=False):
        control = editor.control
        editor.prepare_destroy()
        if remove:
            index = self.find_index_of_editor(editor)
            self.notebook.RemovePage(index)
            control.Destroy()
        control.editor = None
        editor.control = None
        if not quit:
            wx.CallAfter(self.find_active_editor)
        del editor

    def load_file(self, path, current_editor=None, args=None, show_progress_bar=None, default_editor_cls=None):
        log.debug(f"load_file: {path}, current_editor={current_editor}, args={args}")
        try:
            filesystem.filesystem_path(path)
        except FileNotFoundError:
            show_progress_bar = False
        if show_progress_bar is None:
            show_progress_bar = wx.GetApp().show_progress_bar_on_load
        if show_progress_bar:
            # usually True except when loading a file before the Frame is shown
            # (as in app startup). It uses Yield and can cause refresh events
            # to happen before the editor is finished loading.
            progress_log.info(f"START=Loading {path}...")
        try:
            file_metadata = loader.identify_file(path)
            log.debug(f"load_file: file_metadata={file_metadata}")
            if current_editor is not None and current_editor.can_load_file(file_metadata):
                current_editor.load_file(file_metadata)
                return
            document = identify_document(file_metadata)
            if default_editor_cls is not None:
                current_editor = default_editor_cls
            new_editor = self.add_document(document, current_editor, args)
        except Exception as e:
            # force close the progress bar so the error dialog doesn't get lost
            # under the progress bar dialog
            progress_log.info("END")
            import traceback
            traceback.print_exc()
            self.error(str(e))
        else:
            new_editor.load_success(document.uri)
            index = self.find_index_of_editor(new_editor)
            self.notebook.SetPageText(index, new_editor.tab_name)
        finally:
            if show_progress_bar:
                wx.CallAfter(progress_log.info, f"END")
        self.set_title()

    def make_active(self, editor, force=False, in_event_loop=False):
        last = self.active_editor
        self.active_editor = editor
        wx.GetApp().active_frame = self
        if force or last != editor or self.raw_menubar.GetMenuCount() == 0:
            if in_event_loop:
                wx.CallAfter(self.make_active_rebuild, editor)
            else:
                self.make_active_rebuild(editor)

    def make_active_rebuild(self, editor):
        log.debug(f"make_active_rebuild: for {editor}")
        try:
            index = self.find_index_of_control(editor.control)
        except errors.EditorNotFound:
            log.debug("attempt to rebuild removed editor; probably the result of a CallAfter")
        else:
            self.create_menubar()
            self.sync_menubar()
            self.create_toolbar()
            self.sync_toolbar()
            self.create_statusbar()
            self.sync_statusbar()
            self.create_keybindings()
            log.debug(f"setting tab focus to {index}")
            self.notebook.SetSelection(index)
            editor.control.SetFocus()
            self.sync_name()
            self.sync_can_paste()

    def enumerate_tabs(self):
        for index in range(self.notebook.GetPageCount()):
            control = self.notebook.GetPage(index)
            log.debug(f"index={index}, control={control}, editor={control.editor}")
            yield index, control, control.editor

    def close_all_tabs(self):
        editors = list(self.editors)
        for editor in editors:
            log.debug(f"Closing {editor}")
            self.close_editor(editor, quit=True)

    def find_index_of_editor(self, editor):
        return self.find_index_of_control(editor.control)

    def find_editor_from_control(self, control):
        for index in range(self.notebook.GetPageCount()):
            if control == self.notebook.GetPage(index):
                return control.editor
        raise errors.EditorNotFound

    def find_index_of_control(self, control):
        log.debug(f"find_index_of_control: looking for control={control}")
        for index in range(self.notebook.GetPageCount()):
            notebook_control = self.notebook.GetPage(index)
            log.debug(f"find_index_of_control: index={index}, control={notebook_control}")
            if control == notebook_control:
                return index
        raise errors.EditorNotFound

    def find_editor_from_index(self, index):
        control = self.notebook.GetPage(index)
        return self.find_editor_from_control(control)

    def find_active_editor(self):
        # sidebar menu activation from TileManager can cause an activate event,
        # so only process one if the main window is getting activated.
        current_active = wx.GetActiveWindow()
        if current_active is None:
            # MacOS apparently doesn't return valid value, so we have to
            # process it.
            pass
        elif not hasattr(current_active, "make_active"):
            return
        control = self.notebook.GetCurrentPage()
        try:
            editor = self.find_editor_from_control(control)
        except errors.SawxError:
            pass
        else:
            self.make_active(editor, True)

    #### Event callbacks

    def on_menu_open_win(self, evt):
        # windows only works when updating the menu during the event call
        log.debug(f"on_menu_open_win: syncing menubar. From {evt.GetMenu()}")
        self.sync_menubar()
        evt.Skip()

    def on_menu_open_linux(self, evt):
        # workaround for linux which crashes updating the menu bar during an event
        log.debug(f"on_menu_open: syncing menubar.")
        wx.CallAfter(self.sync_menubar)
        evt.Skip()

    def on_menu_open_mac(self, evt):
        # workaround for Mac which sends the EVT_MENU_OPEN for lots of stuff unrelated to menus
        state = wx.GetMouseState()
        if state.LeftIsDown():
            log.debug(f"on_menu_open_mac: syncing menubar. From {evt.GetMenu()}")
            wx.CallAfter(self.sync_menubar)
        else:
            log.debug(f"on_menu_open_mac: skipping menubar sync because mouse isn't down. From {evt.GetMenu()}")
        evt.Skip()

    def on_menu(self, evt):
        action_id = evt.GetId()
        log.debug(f"on_menu: menu id: {action_id}")
        action = None
        try:
            action_key, action = self.menubar.valid_id_map[action_id]
        except KeyError:
            try:
                action_key, action = self.toolbar.valid_id_map[action_id]
            except KeyError:
                log.warning("on_menu: No id {action_id} found in menubar or toolbar")
        if action is not None:
            log.debug(f"on_menu: action_key={action_key}, action={action}")
            try:
                wx.CallAfter(action.perform_as_menu_item, action_key)
            except AttributeError:
                log.debug(f"on_menu: no perform method for {action}")
        evt.Skip()

    def on_page_changed(self, evt):
        index = evt.GetSelection()
        editor = self.find_editor_from_index(index)

        # Can get stuck in a loop if the page changes while you're still in the
        # initial process of displaying the frame.
        still_booting = wx.GetApp().in_bootup_process
        log.debug(f"on_page_changed: page id: {index}, {editor.document.uri}, still_booting={still_booting}")
        self.make_active(editor, force=True, in_event_loop=not still_booting)
        evt.Skip()

    def on_page_closing(self, evt):
        index = evt.GetSelection()
        editor = self.find_editor_from_index(index)
        self.pending_editor_close = index, editor
        log.debug(f"on_page_changing: page id: {index}, {editor.document.uri}")
        editor.prepare_destroy()
        evt.Skip()

    def on_page_closed(self, evt):
        index, editor = self.pending_editor_close
        count = self.notebook.GetPageCount()
        if index >= count:
            index = count - 1
        log.debug(f"on_page_closed: new active page id: {index}")
        wx.CallAfter(self.close_editor, editor, remove=False)
        self.pending_editor_close = None
        evt.Skip()

    def on_activate(self, evt):
        is_active = evt.GetActive()
        log.debug(f"on_activate: {is_active}")
        if is_active:
            wx.CallAfter(self.find_active_editor)
        evt.Skip()

    def on_close(self, evt):
        self.close_all_tabs()
        evt.Skip()

    def on_char_hook(self, evt):
        key_id = (evt.GetModifiers(), evt.GetKeyCode())
        log.debug(f"on_char_hook: key: {key_id}")
        try:
            action_key, action = self.keybindings.valid_key_map[key_id]
            try:
                log.debug(f"found action {action}")
                action.perform_as_keystroke(action_key)
            except errors.ProcessKeystrokeNormally as e:
                log.debug(f"processing keystroke normally instead of action: {e}")
                evt.Skip()
        except KeyError as e:
            log.debug(f"key id {key_id} not found in {self.keybindings.valid_key_map}")
            evt.Skip()

    def on_size(self, evt):
        wx.GetApp().last_window_size = evt.GetSize()
        evt.Skip()


    #### convenience functions for alerts and dialogs

    def prompt_local_file_dialog(self, title="", most_recent=True, save=False, default_filename="", wildcard="*"):
        """Display an "open file" dialog to load from the local filesystem,
        defaulting to the most recently used directory.

        If there is no previously used directory, default to the directory of
        the current file.

        Returns the directory path on the filesystem, or None if not found.
        """
        dirpath = ""
        if save:
            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT
        else:
            style=wx.FD_OPEN | wx.FD_CHANGE_DIR | wx.FD_FILE_MUST_EXIST | wx.FD_PREVIEW

        action = "save as" if save else "open"
        if not title:
            title = "Save File" if save else "Open File"
        if self.active_editor:
            # will try path of current file
            dirpath = self.active_editor.best_file_save_dir

        dlg = wx.FileDialog(self, message=title, defaultFile=default_filename, wildcard=wildcard, style=style)

        if dlg.ShowModal() == wx.ID_OK:
            return dlg.GetPath()
        return None

    def prompt(self, message, title='Prompt', default_value=""):
        """ Convenience method to show a text entry dialog."""
        d = dialogs.SimplePromptDialog(self, message, title, default_value)
        return d.show_and_get_value()

    def confirm(self, message, title="Confirm", cancel=False, yes_default=False, no_label=wx.ID_NO, yes_label=wx.ID_YES):
        """ Convenience method to show a confirmation dialog. """

        style = wx.ICON_INFORMATION | wx.YES_NO
        if cancel:
            style |= wx.CANCEL
        elif yes_default:
            style |= wx.YES_DEFAULT
        else:
            style |= wx.NO_DEFAULT
        dlg = wx.MessageDialog(self, message, title, style)
        dlg.SetYesNoLabels(yes_label, no_label)
        state = dlg.ShowModal()
        dlg.Destroy()
        return None if state == wx.ID_CANCEL else state == wx.ID_YES

    def confirm_cancel(self, message, title="Confirm"):
        """ Convenience method to show a confirmation dialog.

        Returns None if Cancel was chosen in addition to the usual True/False.
        """
        return self.confirm(message, title, cancel=True)

    def information(self, message, title='Information'):
        """ Convenience method to show an information message dialog. """

        current = self.FindFocus()
        dlg = wx.MessageDialog(current, message, title, wx.ICON_INFORMATION | wx.OK)
        dlg.ShowModal()
        dlg.Destroy()

    def warning(self, message, title='Warning'):
        """ Convenience method to show a warning message dialog. """
        current = self.FindFocus()
        dlg = wx.MessageDialog(current, message, title, wx.ICON_WARNING | wx.OK)
        dlg.ShowModal()
        dlg.Destroy()

    def error(self, message, title='Error'):
        """ Convenience method to show an error message dialog. """
        current = self.FindFocus()
        dlg = wx.MessageDialog(current, message, title, wx.ICON_ERROR | wx.OK)
        dlg.ShowModal()
        dlg.Destroy()

    def status_message(self, message, debug=False):
        if debug:
            self.raw_statusbar.debug(message)
        else:
            self.raw_statusbar.message(message)


class FakeNotebook(wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent, -1)
        self.notebook_page = None
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

    def GetPageCount(self):
        return 1 if self.notebook_page is not None else 0

    def GetPage(self, index):
        return self.notebook_page

    def GetCurrentPage(self):
        return self.notebook_page

    def GetSelection(self):
        return 0

    def SetSelection(self, index):
        pass

    def SetPageText(self, index, title):
        pass

    def AddPage(self, control, name):
        if self.notebook_page is None:
            self.notebook_page = control
            self.GetSizer().Add(control, 1, wx.GROW)
        else:
            raise IndexError("Only one page allowed")

    def RemovePage(self, index):
        page = self.notebook_page
        self.notebook_page = None
        self.GetSizer().Clear()
        return page


class SawxTablessFrame(SawxFrame):
    def create_notebook(self):
        notebook = FakeNotebook(self)
        return notebook



import os
import sys
import time
import zipfile

from datetime import datetime
import wx

from .filesystem import fsopen as open
from .filesystem import filesystem_path
from .utils.textutil import guessBinary
from .utils.pyutil import get_plugins

import logging
log = logging.getLogger(__name__)


class FileGuess:
    def __init__(self, uri):
        self.uri = uri
        self.fh = open(uri, 'rb')
        self._sample_data = None
        self._sample_lines = None
        self._all_data = None
        self._is_binary = None
        self._is_zipfile = None
        self._zipfile = None
        self._filesystem_path = None

    @property
    def sample_data(self):
        if self._sample_data is None:
            self.fh.seek(0)
            self._sample_data = self.fh.read(10240)
        return self._sample_data

    @property
    def sample_lines(self):
        if self._sample_lines is None:
            self._sample_lines = self.sample_data.splitlines()
        return self._sample_lines

    @property
    def all_data(self):
        if self._all_data is None:
            self.fh.seek(0)
            self._all_data = self.fh.read()
        return self._all_data

    @property
    def is_binary(self):
        if self._is_binary is None:
            self._is_binary = guessBinary(self.sample_data)
        return self._is_binary

    @property
    def is_text(self):
        return not self.is_binary

    @property
    def is_zipfile(self):
        if self._is_zipfile is None:
            self.fh.seek(0)
            self._is_zipfile = zipfile.is_zipfile(self.fh)
        return self._is_zipfile

    @property
    def zipfile(self):
        if self._zipfile is None:
            self.fh.seek(0)
            self._zipfile = zipfile.ZipFile(self.fh)
        return self._zipfile

    @property
    def filesystem_path(self):
        if self._filesystem_path is None:
            self._filesystem_path = filesystem_path(self.uri)  # may raise OSError
        return self._filesystem_path

    def zipfile_contains(self, filename):
        if not self.is_zipfile:
            return False
        try:
            info = self.zipfile.getinfo(filename)
            log.debug(f"zipfile_contains: {filename}: {info}")
        except KeyError:
            return False
        return True

    def zipfile_contains_extension(self, ext):
        if not self.is_zipfile:
            return False
        for item in self.zipfile.infolist():
            if item.filename.endswith(ext):
                return True
        return True


def identify_file(uri, match_multiple=False):
    """Examine the file to determine MIME type and other salient info to
    allow the loader to chose an editor with which to open the file

    Returns a dict containing (at least) the keys:

    * uri -- the uri of the file
    * mime  -- the text string identifying the MIME type

    and possibly other keys that may be used by specific loaders for specific
    types of data.
    """
    loaders = get_plugins('sawx.loaders')
    log.debug(f"identify_file: identifying file {uri} using {loaders}")
    hits = []
    binary_fallback = None
    text_fallback = None
    file_guess = FileGuess(uri)
    for loader in loaders:
        log.debug(f"identify_file: trying loader {loader}")
        file_metadata = loader.identify_loader(file_guess)
        if file_metadata:
            file_metadata['uri'] = uri
            mime_type = file_metadata['mime']
            if mime_type == "application/octet-stream":
                log.debug(f"identify_file: identified as generic type {mime_type}")
                if not binary_fallback:
                    binary_fallback = file_metadata
            elif mime_type == "text/plain":
                log.debug(f"identify_file: identified as generic type {mime_type}")
                if not text_fallback:
                    text_fallback = file_metadata
            else:
                log.debug(f"identify_file: identified: {file_metadata}")
                if not match_multiple:
                    return file_metadata
                hits.append(file_metadata)

    # how to find best guess?
    if hits:
        log.debug(f"identify_file: found {hits}")
        return hits[0]
    else:
        fallback = text_fallback or binary_fallback  # prefer text match over binary
        if not fallback:
            fallback = dict(mime="application/octet-stream", uri=uri)
        log.debug(f"identify_file: identified only as the generic {fallback}")
        return fallback



import numpy as np


def to_numpy(value):
    if type(value) is np.ndarray:
        return value
    elif type(value) is bytes:
        return np.fromstring(value, dtype=np.uint8)
    elif type(value) is list:
        return np.asarray(value, dtype=np.uint8)
    raise TypeError("Can't convert to numpy data")


def count_in_range(source_list, lowest_valid, highest_valid):
    """Count the number of items in the list between the lowest and highest
    values

    """
    npsource = np.asarray(source_list, dtype=np.uint32)
    count_low = len(np.where(npsource < lowest_valid)[0])
    count_high = len(np.where(npsource > highest_valid)[0])
    return count_low, count_high

# Fast integer-multiple scaling of bitmaps
def intscale(arr, hscale, wscale=None):
    if wscale is None:
        wscale = hscale
    hscale, wscale = int(hscale), int(wscale)
    h, w, depth = arr.shape
    if hscale == 1 and wscale == 1:
        return arr
    elif hscale < 1 or wscale < 1:
        raise ValueError("Scale must be an integer greater than 1")
    elif hscale == wscale:
        # some speedups for special cases where scale value is the same in both
        # directions
        if hscale == 2:
            output = np.empty((h * hscale, w * wscale, depth), dtype=np.uint8)
            # duplicate source pixels into two columns
            output[::2,::2,:] = arr[:,:,:]
            output[::2,1::2,:] = arr[:,:,:]
            # then copy completed rows to the row below
            output[1::2,:,:] = output[0::2,:,:]
            return output
        elif hscale == 3:
            output = np.empty((h * hscale, w * wscale, depth), dtype=np.uint8)
            # duplicate source pixels into three columns
            output[::3,::3,:] = arr[:,:,:]
            output[::3,1::3,:] = arr[:,:,:]
            output[::3,2::3,:] = arr[:,:,:]
            # then copy completed rows to the next two rows
            output[1::3,:,:] = output[::3,:,:]
            output[2::3,:,:] = output[::3,:,:]
            return output
        elif hscale == 4:
            output = np.empty((h * hscale, w * wscale, depth), dtype=np.uint8)
            # duplicate source pixels into four columns
            output[::4,0::4,:] = arr[:,:,:]
            output[::4,1::4,:] = arr[:,:,:]
            output[::4,2::4,:] = arr[:,:,:]
            output[::4,3::4,:] = arr[:,:,:]
            # copy completed rows to the next three rows
            output[1::4,:,:] = output[::4,:,:]
            output[2::4,:,:] = output[::4,:,:]
            output[3::4,:,:] = output[::4,:,:]
            return output
    output = np.empty((h * hscale, w * wscale, depth), dtype=np.uint8)
    # duplicate source pixels horizontally first onto the first line
    for i in range(wscale):
        output[::hscale,i::wscale,:] = arr[:,:,:]
    # duplicate lines to rows below them
    for i in range(1, hscale):
        output[i::hscale,:,:] = output[0::hscale,:,:]

    return output

# Fast integer-multiple scaling of bitmaps in the x direction
def intwscale(arr, scale):
    """Fast integer-multiple scaling of bitmaps.

    The scale is applied to both the width and height. Arrays are assumed to
    have the shape (height, width, depth)
    """
    scale = int(scale)
    if scale == 1:
        return arr
    elif scale < 1:
        raise ValueError("Scale must be an integer greater than 1")
    h, w, depth = arr.shape
    output = np.empty((h, w * scale, depth), dtype=np.uint8)
    if scale == 2:
        # duplicate source pixels into two columns
        output[::,0::2,:] = arr[:,:,:]
        output[::,1::2,:] = arr[:,:,:]
    elif scale == 3:
        # duplicate source pixels into three columns
        output[::,0::3,:] = arr[:,:,:]
        output[::,1::3,:] = arr[:,:,:]
        output[::,2::3,:] = arr[:,:,:]
    elif scale == 4:
        # duplicate source pixels into three columns
        output[::,0::4,:] = arr[:,:,:]
        output[::,1::4,:] = arr[:,:,:]
        output[::,2::4,:] = arr[:,:,:]
        output[::,3::4,:] = arr[:,:,:]
    else:
        raise ValueError("Scale greater than 4 not yet implemented")

    return output

def intwscale_font(arr, scale):
    """Fast integer-multiple scaling of font bitmaps.

    The scale is applied to both the width and height. Arrays are assumed to
    have the shape (num_glyphs, height, width, depth)
    """
    scale = int(scale)
    if scale == 1:
        return arr
    elif scale < 1:
        raise ValueError("Scale must be an integer greater than 1")
    num_glyphs, h, w, depth = arr.shape
    output = np.empty((num_glyphs, h, w * scale, depth), dtype=np.uint8)
    if scale == 2:
        # duplicate source pixels into two columns
        output[:,:,0::2,:] = arr[:,:,:,:]
        output[:,:,1::2,:] = arr[:,:,:,:]
    elif scale == 3:
        # duplicate source pixels into three columns
        output[:,:,0::3,:] = arr[:,:,:,:]
        output[:,:,1::3,:] = arr[:,:,:,:]
        output[:,:,2::3,:] = arr[:,:,:,:]
    elif scale == 4:
        # duplicate source pixels into three columns
        output[:,:,0::4,:] = arr[:,:,:,:]
        output[:,:,1::4,:] = arr[:,:,:,:]
        output[:,:,2::4,:] = arr[:,:,:,:]
        output[:,:,3::4,:] = arr[:,:,:,:]
    else:
        raise ValueError("Scale greater than 4 not yet implemented")

    return output



import os
import sys
import glob
import json
import pkg_resources

import jsonpickle
import jsonpickle.ext.numpy as jsonpickle_numpy
jsonpickle_numpy.register_handlers()
from .ui import fonts  # register wxFont jsonpickle handler

from datetime import datetime

import appdirs
from . import filesystem

import logging
log = logging.getLogger(__name__)


# Custom jsonpickle handlers

import wx

class wxColorHandler(jsonpickle.handlers.BaseHandler):
    def flatten(self, obj, data):
        data["rgba"] = [obj.Red(), obj.Green(), obj.Blue(), obj.Alpha()]
        return data

    def restore(self, obj):
        r, g, b, a = obj["rgba"]
        return wx.Colour(r, g, b, a)

jsonpickle.handlers.registry.register(wx.Colour, wxColorHandler)




config_base_dir = None
log_dir = None
log_file_ext = None
cache_dir = None
user_data_dir = None
template_dir = None

def setup_file_persistence(app_name):
    global config_base_dir, log_dir, log_file_ext, cache_dir, user_data_dir, template_dir

    dirname = appdirs.user_config_dir(app_name)
    config_base_dir = dirname
    template_dir = get_config_subdir("templates")
    filesystem.template_paths[0:0] = [template_dir]

    # Make sure it exists!
    if not os.path.exists(config_base_dir):
        os.makedirs(config_base_dir)

    dirname = appdirs.user_log_dir(app_name)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    log_dir = dirname

    log_file_ext = "-%s" % datetime.now().strftime("%Y%m%d-%H%M%S")

    dirname = appdirs.user_cache_dir(app_name)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    cache_dir = dirname

    dirname = appdirs.user_data_dir(app_name)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    user_data_dir = dirname

    # Prevent py2exe from creating a dialog box on exit saying that there
    # are error messages.  It thinks that anything written to stderr is an
    # error, and the python logging module redirects everything to stderr.
    # Instead, redirect stderr to a log file in the user log directory
    frozen = getattr(sys, 'frozen', False)
    if frozen in ('dll', 'windows_exe', 'console_exe'):
        # redirect py2exe stderr/stdout to log file
        log = get_log_file_name("py2exe")
        oldlog = sys.stdout
        sys.stdout = open(log, 'w')
        if hasattr(oldlog, "saved_text"):
            sys.stdout.write("".join(oldlog.saved_text))
        sys.stderr = sys.stdout

        # The logging module won't redirect to the new stderr without help
        handler = logging.StreamHandler(sys.stderr)
        logger = logging.getLogger('')
        logger.addHandler(handler)
    else:
        log = get_log_file_name("log")
        handler = logging.FileHandler(log)
        formatter = logging.Formatter("%(levelname)s:%(name)s:%(msg)s")
        handler.setFormatter(formatter)
        logger = logging.getLogger('')
        logger.addHandler(handler)

def get_log_file_name(log_file_name_base, ext=""):
    filename = log_file_name_base + log_file_ext
    if ext:
        if not ext.startswith("."):
            filename += "."
        filename += ext
    else:
        filename += ".log"
    filename = os.path.join(log_dir, filename)
    return filename

def save_log(text, log_file_name_base, ext=""):
    filename = get_log_file_name(log_file_name_base, ext)

    try:
        with open(filename, "w") as fh:
            fh.write(text)
    except IOError:
        log.error("Failed writing %s to %s" % (log_file_name_base, filename))

def get_config_subdir(subdir):
    dirname = os.path.join(config_base_dir, subdir)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    return dirname

def get_config_dir_filename(storage_type, json_name):
    return os.path.join(config_base_dir, json_name + "." + storage_type)

def get_file_config_data(subdir, filename, default_on_error=None, mode='r'):
    try:
        file_path = get_config_dir_filename(subdir, filename)
        with open(file_path, mode) as fh:
            data = fh.read()
        return data
    except IOError:
        # file not found
        return default_on_error

def save_file_config_data(subdir, filename, data, mode='w'):
    file_path = get_config_dir_filename(subdir, filename)
    with open(file_path, mode) as fh:
        fh.write(data)
    return file_path

def get_json_data(json_name, default_on_error=None):
    raw = get_file_config_data("json", json_name)
    if raw is None:
        return default_on_error
    try:
        decoded = jsonpickle.decode(raw)
        try:
            # legacy format is a list with a format identifier as the first
            # entry
            if decoded[0] == "format=v2":
                decoded = decoded[1]
        except KeyError:
            pass
        return decoded
    except ValueError:
        # bad JSON format
        log.error("Bad JSON format in preferences file: %s" % json_name)
        return default_on_error

def save_json_data(json_name, data):
    encoded = jsonpickle.encode(data)
    return save_file_config_data("json", json_name, encoded)

def get_bson_data(bson_name):
    import bson

    raw = get_file_config_data("bson", bson_name, mode='rb')
    if raw is not None and len(raw) > 0:
        bson_data = bson.loads(raw)
        data = bson_data[bson_name]
    else:
        raise IOError("Blank BSON data")
    return data

def save_bson_data(bson_name, data):
    import bson

    bson_data = {bson_name: data}
    raw = bson.dumps(bson_data)
    save_file_config_data("bson", bson_name, raw, mode='wb')

def get_user_dir(subdir):
    dirname = os.path.join(user_data_dir, subdir)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    return dirname

def get_user_dir_filename(subdir, json_name):
    return os.path.join(get_user_dir(subdir), json_name)

def get_user_data(subdir, filename, default_on_error=None, mode='r'):
    try:
        file_path = get_user_dir_filename(subdir, filename)
        with open(file_path, mode) as fh:
            data = fh.read()
        return data
    except IOError:
        # file not found
        return default_on_error

def save_user_data(subdir, filename, data, mode='w'):
    file_path = get_user_dir_filename(subdir, filename)
    with open(file_path, mode) as fh:
        fh.write(data)

def get_available_user_data(subdir):
    config_dir = get_user_dir(subdir)
    globname = os.path.join(config_dir, "*")
    available = [os.path.basename(a) for a in glob.glob(globname)]
    return available

def get_text_user_data(subdir, filename, default_on_error=None):
    return get_user_data(subdir, filename, default_on_error, 'r')

def save_text_user_data(subdir, filename, data):
    return save_user_data(subdir, filename, data)

def get_binary_user_data(subdir, filename, default_on_error=None):
    return get_user_data(subdir, filename, default_on_error, 'rb')

def save_binary_user_data(subdir, filename, data):
    return save_user_data(subdir, filename, data, 'wb')

def get_cache_dir(subdir):
    dirname = os.path.join(cache_dir, subdir)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    return dirname


# Templates

def get_template_fh(name, include_user_defined=True):
    # resource path will point to the omnivore/templates directory
    pathname = filesystem.find_template_path(name, include_user_defined)
    log.debug("Loading template for %s: %s" % (name, pathname))
    fh = filesystem.fsopen(pathname, "rb")
    return fh

def get_template(name, include_user_defined=True):
    # resource path will point to the omnivore/templates directory
    fh = get_template_fh(name, include_user_defined)
    source = fh.read()
    return source

def save_template(name, contents, binary=True):
    pathname = os.path.normpath(os.path.join(template_dir, name))
    if binary:
        mode = "wb"
    else:
        mode = "w"
    fh = filesystem.fsopen(pathname, mode)
    fh.write(contents)

class TemplateItem(dict):
    def __init__(self, data_file_path, keyword, inf_type, json_dict):
        dict.__init__(self)
        self.data_file_path = data_file_path
        self.keyword = keyword
        self.inf_type = inf_type
        self._data_bytes = None
        for k, v in json_dict.items():
            self[k] = v
            setattr(self, k, v)

    def __str__(self):
        return self.get("label", "") or self.get("description", "") or self.get("name", "") or self.keyword

    def __lt__(self, other):
        return str(self) < str(other)

    @property
    def data_bytes(self):
        if not self._data_bytes:
            with open(self.data_file_path, 'rb') as fh:
                self._data_bytes = fh.read()
        return self._data_bytes


def iter_templates(inf_type=None):
    """Returns dictionary containing info about specific template types

    Dict contains keys:

    * pathname  - the full pathname to the template to differentiate default
      and user templates with the same name
    * uri
    * task  - editor name
    """
    templates = {}
    for template_path in filesystem.glob_in_paths(filesystem.template_paths):
        if template_path.endswith(".inf"):
            # skip loading inf files
            continue
        inf_path = template_path + ".inf"
        j = None
        if os.path.exists(template_path) and os.path.isfile(template_path):
            log.debug(f"checking template {template_path} for {inf_type}")
            keyword = os.path.basename(template_path)
            keyword2, ext = os.path.splitext(keyword)
            if ext and ext[1:] == inf_type:
                keyword = keyword2
                ext = inf_type
            if os.path.exists(inf_path):
                try:
                    log.debug("Loading json file %s" % inf_path)
                    with open(inf_path, "r") as fh:
                        j = json.loads(fh.read())
                except ValueError as e:
                    log.error(f"Template properties error in {inf_path}: {e}")
                    continue
                if inf_type:
                    if "type" not in j and ext and ext != inf_type:
                        continue
                    elif "type" in j and j["type"] != inf_type:
                        continue
                j["pathname"] = template_path
                j["uri"] = "template://" + template_path
                if "task" in j and j["task"] == "hex_edit":
                    j["task"] = "byte_edit"
            elif ext == inf_type:
                j = {}
                j["type"] = inf_type
                j["pathname"] = template_path
                j["uri"] = "template://" + template_path
            if j is not None:
                item = TemplateItem(template_path, keyword, inf_type, j)
                log.debug(f"template: {item}")
                yield item


# Remember hooks

def restore_from_last_time():
    modules = []
    for entry_point in pkg_resources.iter_entry_points('sawx.remember'):
        try:
            mod = entry_point.load()
        except Exception as e:
            log.error(f"Failed importing remember entry point {entry_point.name}: {e}")
            import traceback
            traceback.print_exc()
        else:
            log.debug(f"restore_from_last_time: Found module {entry_point.name}")
            try:
                restore = getattr(mod, 'restore_from_last_time')
            except AttributeError:
                log.warning(f"restore_from_last_time: no restore function in module {entry_point.name}")
            else:
                restore()
                modules.append(mod)
    return modules


def remember_for_next_time(modules):
    for mod in modules:
        log.debug(f"remember_for_next_time: {mod}")
        try:
            remember = getattr(mod, 'remember_for_next_time')
        except AttributeError:
            log.warning(f"remember_for_next_time: no remember function in module {mod}")
        else:
            remember()



import os

from ..document import SawxDocument
from ..filesystem import fsopen as open

import logging
log = logging.getLogger(__name__)


class TextDocument(SawxDocument):
    def load_raw_data(self):
        fh = open(self.uri, 'r')
        return fh.read()

    def calc_raw_data(self, raw):
        return str(raw)

    def calc_raw_data_to_save(self):
        return self.raw_data

    def save_raw_data(self, uri, raw_data):
        fh = open(uri, 'w')
        log.debug("saving to %s" % uri)
        fh.write(raw_data)
        fh.close()

    # won't automatically match anything; must force this editor with the -t
    # command line flag
    @classmethod
    def can_load_file_exact(cls, file_metadata):
        return False

    @classmethod
    def can_load_file_generic(cls, file_metadata):
        return file_metadata['mime'].startswith("text/")



import collections

import wx

from .action import get_action_id
from . import errors

import logging
log = logging.getLogger("menubar")
sync_log = logging.getLogger("sync-menubar")


def expand(action_key, editor):
    if str(action_key) == action_key and "(" in action_key and action_key.endswith(")"):
        func_name, args_string = action_key[:-1].split("(", 1)
        func = getattr(editor, func_name)
        args = args_string.split(",")
        if args[0] == "":
            expansion = func()
        else:
            expansion = func(*args)
        log.debug(f"expanding {action_key} to {expansion}")
        for item in expansion:
            yield item
    else:
        yield action_key


class MenuDescription:
    def __init__(self, menu_name, desc, editor, valid_id_map, keybinding_desc=None, popup_data=None):
        self.menu = wx.Menu()
        self.name = menu_name
        log.debug(f"adding menu {desc}")
        allow_separator = False
        if keybinding_desc is None:
            keybinding_desc = {}
        for action_key in desc:
            for action_key in expand(action_key, editor):
                if action_key is None:
                    if allow_separator:
                        self.menu.AppendSeparator()
                        allow_separator = False
                elif str(action_key) == action_key:
                    if action_key.startswith("-"):
                        if allow_separator:
                            self.menu.AppendSeparator()
                            allow_separator = False
                    else:
                        # usable_actions limit the visible actions to what the current editor supports
                        try:
                            action = editor.calc_usable_action(action_key)
                        except KeyError as e:
                            log.debug(f"action {action_key} not used in this editor: {e}")
                            pass
                        else:
                            action.popup_data = popup_data
                            # a single action can create multiple entries
                            try:
                                action_keys = action.calc_menu_sub_keys(editor)
                                log.debug(f"action {action_key} created subkeys {action_keys}")
                            except AttributeError:
                                action_keys = [action_key]
                            for action_key in action_keys:
                                id = get_action_id(action_key)
                                valid_id_map[id] = (action_key, action)
                                action.append_to_menu(self.menu, id, action_key)
                                try:
                                    accel_text = keybinding_desc[action_key]
                                except KeyError:
                                    pass
                                else:
                                    if accel_text:
                                        accel_entry = wx.AcceleratorEntry()
                                        accel_entry.FromString(accel_text)
                                        menu_item = self.menu.FindItemById(id)
                                        try:
                                            menu_item.SetAccel(accel_entry)
                                        except Exception:
                                            log.error(f"Invalid key binding {accel_text} for {action_key}")
                                allow_separator = True
                else:
                    submenu_desc = action_key
                    submenu_name = submenu_desc[0]
                    submenu = MenuDescription(submenu_name, submenu_desc[1:], editor, valid_id_map, keybinding_desc, popup_data)
                    if submenu.count > 0:
                        self.menu.AppendSubMenu(submenu.menu, submenu.name)
                        allow_separator = True

    @property
    def count(self):
        return self.menu.GetMenuItemCount()

    def sync_with_editor(self, valid_id_map):
        sync_with_editor(valid_id_map, self.menu)


class MenubarDescription:
    def __init__(self, parent, editor):
        self.menus = []
        self.valid_id_map = collections.OrderedDict()
        num_old_menus = parent.raw_menubar.GetMenuCount()
        num_new_menus = 0
        for desc in editor.menubar_desc:
            menu_name = desc[0]
            menu = MenuDescription(menu_name, desc[1:], editor, self.valid_id_map, editor.keybinding_desc)
            if menu.count > 0:
                if num_new_menus < num_old_menus:
                    parent.raw_menubar.Replace(num_new_menus, menu.menu, menu.name)
                else:
                    parent.raw_menubar.Append(menu.menu, menu.name)
                self.menus.append(menu)
                num_new_menus += 1
        while num_new_menus < num_old_menus:
            parent.raw_menubar.Remove(num_new_menus)
            num_old_menus -= 1

    def sync_with_editor(self, menubar_control):
        sync_with_editor(self.valid_id_map, menubar_control)


def sync_with_editor(valid_id_map, control):
    for id, (action_key, action) in valid_id_map.items():
        sync_log.debug(f"syncing {id}: {action_key}, {action}")
        menu_item = control.FindItemById(id)
        try:
            action.sync_menu_item_from_editor(action_key, menu_item)
        except AttributeError as e:
            sync_log.debug(f"Skipping sync of {action_key} menu item from {action.editor}: {e}")
        except errors.RecreateDynamicMenuBar:
            raise
        except:
            log.error(f"Error in sync_menu_item_from_editor for {action_key}")
            raise



import os

from .. import filesystem
from .. import errors

import logging
log = logging.getLogger(__name__)


def get_htmlhelp(name):
    attempts = []
    for subdir in filesystem.help_paths:
        pathname = os.path.normpath(os.path.join(subdir, name))
        log.debug("Checking for htmlhelp at %s" % pathname)
        if os.path.exists(pathname):
            log.debug("Found htmlhelp for %s: %s" % (name, pathname))
            return pathname
        attempts.append(pathname)
    else:
        log.debug("No htmlhelp found for %s in %s" % (name, attempts))
        raise errors.MissingDocumentationError("Unable to locate %s help files; installation error?\nThe files should be in one of the following:\n%s\n\nbut were not found." % (name, "\n  * " + "\n  * ".join([os.path.dirname(d) for d in attempts])))



import logging
log = logging.getLogger(__name__)


def identify_loader(file_guess):
    mime_type = None
    if not file_guess.is_binary:
        header = file_guess.sample_data.strip().lower()
        if header.startswith(b"<"):
            if header.startswith(b"<?xml"):
                found_xml = True
            else:
                found_xml = False
            if b"<rss" in header:
                mime_type = "application/rss+xml"
            elif b"<!doctype html" in header or b"<html" in header:
                mime_type = "text/html"
            elif found_xml:
                mime_type = "text/xml"
        elif header.startswith(b"#!"):
            line = header[2:80].lower().strip()
            if line.startswith(b"/usr/bin/env"):
                line = line[12:].strip()
            words = line.split()
            names = words[0].split(b"/")
            if names[-1]:
                mime_type = "text/%s" % names[-1].decode("utf-8")

        if not mime_type:
            mime_type = "text/plain"

    if mime_type:
        log.debug(f"text loader: identified {mime_type}")
        return dict(mime=mime_type, ext="")
    else:
        log.debug(f"text loader: unidentified")
        return None



from ._metadata import __author__, __author_email__, __url__, __download_url__, __bug_report_url__

try:
    from ._version import __version__
except ImportError:
    __version__ = "dev"

from .application import SawxApp
from .frame import SawxFrame
from .editor import SawxEditor
from .action import SawxAction, SawxActionRadioMixin
from . import errors

import logging
log = logging.getLogger(__name__)



""" Download manager using background threads
"""
import os
import time
import urllib.request, urllib.error, urllib.parse
import ssl
import tempfile

import wx
import wx.lib.scrolledpanel as scrolled

from ..utils.background_http import BaseRequest, BackgroundHttpMultiDownloader

import logging
log = logging.getLogger(__name__)


class NoCallback(object):
    def __call__(self):
        def no_callback(req):
            print(("no callback for", req))
            return
        return no_callback


class DownloadURLRequest(BaseRequest):
    blocksize = 64 * 1024

    debug = False

    def __init__(self, url, path, threadsafe_progress_callback=None, finished_callback=None):
        BaseRequest.__init__(self)
        self.url = url
        self.is_skippable = False
        self.expected_size = 0
        self.size = 0
        self.wants_cancel = False
        self.is_cancelled = False
        self.path = path
        self._threadsafe_progress_callback = None
        self._finished_callback = None
        self.threadsafe_progress_callback = threadsafe_progress_callback
        self.finished_callback = finished_callback

    @property
    def threadsafe_progress_callback(self):
        return self._threadsafe_progress_callback

    @threadsafe_progress_callback.setter
    def threadsafe_progress_callback(self, callback):
        if callback is not None:
            self._threadsafe_progress_callback = callback
        else:
            self._threadsafe_progress_callback = NoCallback()()

    @property
    def finished_callback(self):
        return self._finished_callback

    @finished_callback.setter
    def finished_callback(self, callback):
        if callback is not None:
            self._finished_callback = callback
        else:
            self._finished_callback = NoCallback()()

    def __str__(self):
        if not self.is_started:
            return "%s download pending" % (self.url)
        elif self.is_finished:
            if self.error is None:
                if self.data:
                    return "%d bytes in %s" % (self.size, self.path)
                elif self.is_cancelled:
                    return "%s cancelled, %d of %d bytes" % (self.url, self.size, self.expected_size)
                else:
                    return "%s incomplete download, %d of %d bytes" % (self.url, self.size, self.expected_size)
            else:
                return "%s error: %s" % (self.url, self.error)
        else:
            return "%s downloading, %d/%d" % (self.url, self.size, self.expected_size)

    def cancel(self):
        self.wants_cancel = True

    def get_data_from_server(self):
        try:
            request = urllib.request.Request(self.url)
            context = ssl._create_unverified_context()
            response = urllib.request.urlopen(request, context=context)
            headers = response.info()
            if "Content-Length" in headers:
                self.expected_size = int(headers['Content-Length'])
            finished = False
            tmp_path = self.path + ".download"
            with open(tmp_path, 'wb') as fh:
                while True:
                    chunk = response.read(self.blocksize)
                    if not chunk:
                        finished = True
                        break
                    if self.wants_cancel:
                        self.is_cancelled = True
                        break
                    fh.write(chunk)
                    self.size += len(chunk)
                    self.threadsafe_progress_callback(self)
                    if self.debug:
                        time.sleep(.1)
            if finished:
                try:
                    os.remove(self.path)
                except OSError:
                    pass
                os.rename(tmp_path, self.path)
                self.data = self.path
            self.finish()

        except (urllib.error.URLError, ValueError, OSError) as e:
            self.handle_error(e)

    def handle_error(self, e):
        self.error = e
        self.finish()

    def finish(self):
        self.is_finished = True
        self.threadsafe_progress_callback(self)
        self.threadsafe_progress_callback = None
        wx.CallAfter(self.finished_callback, self, self.error)
        self.finished_callback = None

    def get_gauge_value(self):
        if self.expected_size == 0:
            return -1
        return self.size * 100 / self.expected_size


class RequestStatusControl(wx.Panel):
    border = 5

    def __init__(self, parent, req, **kwargs):
        wx.Panel.__init__(self, parent, -1, **kwargs)
        self.req = req
        hbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(hbox)
        self.text = wx.StaticText(self, -1, req.url, style=wx.ST_ELLIPSIZE_START)
        hbox.Add(self.text, 0, flag=wx.EXPAND|wx.ALL, border=self.border)
        vbox = wx.BoxSizer(wx.HORIZONTAL)
        self.gauge = wx.Gauge(self, -1)
        self.gauge.SetRange(100)
        vbox.Add(self.gauge, 1, flag=wx.EXPAND|wx.ALL)
        self.cancel = wx.Button(self, -1, "Cancel")
        self.cancel.Bind(wx.EVT_BUTTON, self.on_cancel)
        vbox.Add(self.cancel, 0, flag=wx.EXPAND|wx.LEFT, border=self.border)
        hbox.Add(vbox, 0, flag=wx.EXPAND|wx.ALL, border=self.border)
        hbox.Fit(self)
        self.Layout()

    def on_cancel(self, evt):
        self.req.cancel()

    def update(self):
        self.text.SetLabel(str(self.req))
        perc = self.req.get_gauge_value()
        if perc == -1:
            self.gauge.Pulse()
        else:
            self.gauge.SetValue(perc)
        if self.req.is_finished:
            self.cancel.Enable(False)


class DownloadControl(scrolled.ScrolledPanel):
    """
    View of list of downloaded items
    """

    def __init__(self, parent, downloader, path=None, prefix="downloads_", **kwargs):
        scrolled.ScrolledPanel.__init__(self, parent, -1, **kwargs)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.ShowScrollbars(wx.SHOW_SB_NEVER, wx.SHOW_SB_ALWAYS)
        self.SetupScrolling(scroll_x=False)

        hbox = wx.BoxSizer(wx.HORIZONTAL)
        self.header = wx.StaticText(self, -1, "")
        hbox.Add(self.header, 1, flag=wx.ALIGN_CENTER)

        # Adding the button prevents the window from being dismissed when in a
        # sidebar popup, so need to use the lose_focus_helper
        self.lose_focus_helper_function = None

        self.setdir = wx.Button(self, -1, "Download Dir")
        self.setdir.Bind(wx.EVT_BUTTON, self.on_setdir)
        self.setdir.Bind(wx.EVT_KILL_FOCUS, self.on_lose_child_focus)  # helper!
        hbox.Add(self.setdir, 0, flag=wx.EXPAND|wx.ALL)
        sizer.Add(hbox, 0, flag=wx.EXPAND)

        self.SetSizer(sizer)
        sizer.Layout()
        self.Fit()

        self.default_path = path
        self.dir_prefix = prefix
        self.downloader = downloader
        self.req_map = {}
        self.update_header()

    @property
    def path(self):
        if self.default_path:
            return default_path
        return tempfile.mkdtemp(prefix=self.dir_prefix)

    @path.setter
    def path(self, value):
        if value:
            self.default_path = value

    @property
    def num_active(self):
        count = 0
        for req in list(self.req_map.keys()):
            if not req.is_finished:
                count += 1
        return count

    def update_header(self):
        count = self.num_active
        if count == 0:
            text = "No active downloads"
        elif count == 1:
            text = "1 active download"
        else:
            text = "1 active download, %d queued" % (count - 1)
        self.header.SetLabel(text)

    def request_download(self, url, filename, callback):
        if not os.path.isabs(filename):
            filename = os.path.normpath(os.path.join(self.path, filename))
        log.debug("request_download: %s" % filename)
        req = DownloadURLRequest(url, filename, finished_callback=callback)
        self.add_request(req)
        return req

    def add_request(self, req):
        rc = RequestStatusControl(self, req)
        req.threadsafe_progress_callback = self.threadsafe_progress_callback
        self.req_map[req] = rc
        sizer = self.GetSizer()
        sizer.Add(rc, 0, flag=wx.EXPAND)
        self.Layout()
        self.SetupScrolling()
        self.downloader.send_request(req)
        rc.cancel.Bind(wx.EVT_KILL_FOCUS, self.on_lose_child_focus)

    def threadsafe_progress_callback(self, req):
        rc = self.req_map[req]
        wx.CallAfter(rc.update)
        wx.CallAfter(self.update_header)

    def on_setdir(self, evt):
        dlg = wx.DirDialog(self, "Choose the download directory:", style=wx.DD_DEFAULT_STYLE)
        dlg.SetPath(self.path)
        if dlg.ShowModal() == wx.ID_OK:
            self.path = dlg.GetPath()
        dlg.Destroy()

    def on_lose_child_focus(self, evt):
        log.debug("on_lose_child_focus! currently focused: %s next focused: %s" % (self.FindFocus(), evt.GetWindow()))
        evt.Skip()
        if self.lose_focus_helper_function is not None:
            self.lose_focus_helper_function(evt)


if __name__ == "__main__":
    # Due to the package import in the parent directory, running from this
    # directory won't work. Have to hack it:
    #
    # PYTHONPATH=../../.. python download_manager.py
    from ..background_http import BackgroundHttpDownloader
    import wx.lib.inspection

    class MyFrame(wx.Frame):
        def __init__(self, parent, id, title):
            wx.Frame.__init__(self, parent, id, title, wx.DefaultPosition, wx.DefaultSize)
            self.dlcontrol = DownloadControl(self, None)

    class MyApp(wx.App):
        def OnInit(self):
            self.dlframe = MyFrame(None, -1, 'Download Manager')
            self.dlframe.Show(True)
            self.dlframe.Centre()
            return True

    def finished_callback(req, error):
        print(("FINISHED!", req))

    def do_download(dlc):
        req = dlc.request_download('http://playermissile.com', "index.html", finished_callback)
        req2 = dlc.request_download('http://playermissile.com/mame', "mame-index.html", finished_callback)
        req3 = dlc.request_download('http://playermissile.com/jumpman', "jumpman-index.html", finished_callback)
        req4 = dlc.request_download('http://playermissile.com/jumpman', "jumpman-index.html", finished_callback)
        req5 = dlc.request_download('http://playermissile.com/jumpman', "jumpman-index.html", finished_callback)

    app = MyApp(0)
    downloader = BackgroundHttpDownloader()
    app.dlframe.dlcontrol.downloader = downloader
    dlc = app.dlframe.dlcontrol
    DownloadURLRequest.blocksize = 1024
    wx.CallAfter(do_download, dlc)
    inspect = wx.lib.inspection.InspectionTool()
    wx.CallAfter(inspect.Show)

    app.MainLoop()

    downloader = None



import numpy as np

import logging
log = logging.getLogger(__name__)


def collapse_overlapping_ranges(ranges):
    """ Collapse the list of (possibly overlapping) selected ranges into
    a monotonically increasing set of non-overlapping ranges
    """
    opt = []
    start, end = None, None
    for next_start, next_end in sorted(ranges):
        if next_start > next_end:
            next_start, next_end = next_end, next_start
        if start is None:
            start, end = next_start, next_end
        else:
            if next_start > end:
                opt.append((start, end))
                start, end = next_start, next_end
            elif next_end > end:
                end = next_end
    if start is not None:
        opt.append((start, end))
    return opt


def ranges_to_indexes(ranges):
    if len(ranges) == 0:
        return np.zeros([0], dtype=np.uint32)
    return np.hstack((np.arange(r[0], r[1], dtype=np.uint32) for r in ranges))


def indexes_to_ranges(indexes):
    groups = np.split(indexes, np.where(np.diff(indexes) != 1)[0] + 1)
    ranges = []
    for group in groups:
        if np.alen(group) > 0:
            ranges.append((int(group[0]), int(group[-1]) + 1))
    return ranges


def invert_ranges(ranges, last):
    """ Invert the list of (possibly overlapping) selected ranges into a
    monotonically increasing set of non-overlapping ranges that represents the
    opposite of the listed ranges
    """
    # get a monotonically increasing list
    ranges = collapse_overlapping_ranges(ranges)
    inverted = []

    first = 0
    for start, end in ranges:
        if start > first:
            inverted.append((first, start))
            first = end
        else:
            first = end
    if first < last:
        inverted.append((first, last))
    return inverted


def rect_ranges_to_indexes(row_width, start_offset, ranges):
    # Loop over each range to determine the indexes of the selected bytes
    # Returns a unique, monotonically increasing list of indexes to guarantee
    # that each index appears in the list only once.
    if len(ranges) == 0:
        return np.zeros([0], dtype=np.uint32)
    indexes = np.empty([0], dtype=np.uint32)
    log.debug("starting rects: %s" % np.vectorize(hex)(ranges))
    for start, end in ranges:
        if start is None:
            continue
        if start > end:
            start, end = end, start
        end -= 1  # instead of slice format, last byte becomes inclusive
        r1, c1 = divmod(start, row_width)
        r2, c2 = divmod(end, row_width)
        log.debug("before: %x-%x, (%d,%d) -> (%d,%d)" % (start, end, r1, c1, r2, c2))
        if c2 < c1:
            first_row_column_zero = start - c1
            c1, c2 = c2, c1
        else:
            first_row_column_zero = start - c1
        num_cols = c2 - c1 + 1
        num_rows = r2 - r1 + 1
        log.debug("range: %x-%x, (%d,%d) -> (%d,%d), nr=%d nc=%d zeroc=%x" % (start, end, r1, c1, r2, c2, num_rows, num_cols, first_row_column_zero))
        rect_indexes = np.hstack((np.arange(i + c1 + first_row_column_zero, i + c1 + first_row_column_zero + num_cols, dtype=np.uint32) for i in range(0, num_rows * row_width, row_width)))
        indexes = np.hstack((indexes, rect_indexes))

    unique_indexes = np.unique(indexes)
    log.debug("rect indexes: %s" % np.vectorize(hex)(unique_indexes))
    return unique_indexes


def invert_rects(rects, numr, numc):
    # Purely heuristic approach.  An algorithmic approach might be based on:
    # http://stackoverflow.com/questions/30818645 but this one breaks up the
    # entire space into a set of rectangles with boundaries at every possible
    # row/col boundary of all the contained rectangles.

    # Get list of all referenced rows and cols
    rows = sorted(list(set([r[0] for rect in rects for r in rect]).union(set([0, numr]))))
    cols = sorted(list(set([r[1] for rect in rects for r in rect]).union(set([0, numc]))))

    # create inside/outside flags for each row/col in the entire grid
    inside = np.zeros((numr, numc), dtype=np.bool_)
    for [(r1, c1), (r2, c2)] in rects:
        inside[r1:r2, c1:c2] = True

    # create lo/hi values for each subdivision of the grid
    rowpairs = [(rows[i], rows[i+1]) for i in range(len(rows) - 1)]
    colpairs = [(cols[i], cols[i+1]) for i in range(len(cols) - 1)]

    # create a rectangle at each intersection point and if it's outside the
    # original set of rectangles, add it to the list
    outside = []
    for r1, r2 in rowpairs:
        for c1, c2 in colpairs:
            if not inside[r1, c1]:
                outside.append([(r1, c1), (r2, c2)])

    # optimization: merge neighboring rectangles that share common edges

    return outside



""" Threaded URL loading from

http://stackoverflow.com/questions/3817505/wxpython-htmlwindow-freezes-when-loading-images
"""

import urllib.request
import urllib.error
import os
import time
import threading
import queue

import logging
log = logging.getLogger(__name__)


class BaseRequest(object):
    def __init__(self):
        self.url = "no url"
        self.data = None
        self.error = None
        self.is_started = False
        self.is_finished = False
        self.is_skippable = True

    def __str__(self):
        if self.data is None:
            if self.error is None:
                return str(self.url)
            else:
                return "%s returned error: %s" % (self.url, self.error)
        else:
            return "%s returned %d bytes" % (self.url, len(self.data))

    def has_error(self):
        return self.error is not None

    def get_data_using_thread(self):
        self.is_started = True
        self.get_data_from_server()
        self.is_finished = True

    def get_data_from_server(self):
        self.data = "testing..."


class UnskippableRequest(BaseRequest):
    def __init__(self):
        BaseRequest.__init__(self)
        self.is_skippable = False


class URLRequest(BaseRequest):
    def __init__(self, url):
        BaseRequest.__init__(self)
        self.url = url

    def get_data_from_server(self):
        try:
            request = urllib.request.Request(self.url)
            response = urllib.request.urlopen(request)
            self.data = response.read()
        except urllib.error.URLError as e:
            self.error = e


class UnskippableURLRequest(URLRequest):
    def __init__(self, url):
        URLRequest.__init__(self, url)
        self.is_skippable = False


class HttpThread(threading.Thread):
    http_thread_count = 0

    def __init__(self, in_q, out_q, name_prefix="HttpThread"):
        self.__class__.http_thread_count += 1
        threading.Thread.__init__(self, name="%s-%d" % (name_prefix, self.http_thread_count))
        self.in_q = in_q
        self.out_q = out_q

    def get_next(self):
        req = self.in_q.get(True) # blocking
        return req

    def run(self):
        log.debug("%s: starting http thread..." % self.name)
        while True:
            log.debug("%s: waiting for request..." % self.name)
            req = self.get_next()
            if req is None:
                break

            log.debug("%s: loading from %s" % (self.name, req))
            req.get_data_using_thread()
            log.debug("%s: result from %s" % (self.name, req))
            self.out_q.put(req)


class OnlyLatestHttpThread(HttpThread):
    def get_next(self):
        """Return only the latest URL, skip any older ones as being outdated
        
        """
        req = BaseRequest()  # can't use None because None means quit
        wait = True
        while req is not None and req.is_skippable:
            log.debug("skipping req %s, skippable=%s", req, req.is_skippable)
            try:
                req = self.in_q.get(wait)
                log.debug("found req %s", req)
            except queue.Empty:
                break
            wait = False
        return req


class BackgroundHttpDownloader(object):
    def __init__(self):
        self.requests = queue.Queue()
        self.results = queue.Queue()
        self.thread = OnlyLatestHttpThread(self.requests, self.results, "BackgroundHttpDownloader")
        self.thread.start()
        log.debug("Created thread %s" % self.thread.name)
        self.get_server_config()

    def __del__(self):
        self.stop_threads()

    def stop_threads(self):
        self.requests.put(None)
        self.thread.join()
        log.debug("Stopped BackgroundHttpDownloader thread")

    def get_server_config(self):
        pass

    def send_request(self, req):
        self.requests.put(req)

    def get_finished(self):
        finished = []
        try:
            while True:
                data = self.results.get(False)
                finished.append(data)
        except queue.Empty:
            pass
        return finished


class BackgroundHttpMultiDownloader(object):
    def __init__(self, num_workers=4):
        self.requests = queue.Queue()
        self.results = queue.Queue()
        self.threads = []
        for i in range(num_workers):
            thread = HttpThread(self.requests, self.results, "BackgroundHttpMultiDownloader")
            thread.start()
            log.debug("Created thread %s" % thread.name)
            self.threads.append(thread)
        self.get_server_config()

    def __del__(self):
        self.stop_threads()

    def stop_threads(self):
        for t in self.threads:
            self.requests.put(None)
        for t in self.threads:
            t.join()

    def get_server_config(self):
        pass

    def send_request(self, req):
        self.requests.put(req)

    def get_finished(self):
        finished = []
        try:
            while True:
                data = self.results.get(False)
                finished.append(data)
        except queue.Empty:
            pass
        return finished


if __name__ == "__main__":

    downloader = BackgroundHttpDownloader()
    downloader.send_request(URLRequest('http://www.python.org/'))
    downloader.send_request(URLRequest('http://www.doughellmann.com/PyMOTW/contents.html'))
    downloader.send_request(URLRequest('http://playermissile.com'))
    downloader.send_request(URLRequest('http://docs.python.org/release/2.6.8/_static/py.png'))
    downloader.send_request(URLRequest('http://image.tmdb.org/t/p/w342/vpk4hLyiuI2SqCss0T3jeoYcL8E.jpg'))
    downloader.send_request(URLRequest('hvvttp://playermissile.com'))
    first = True
    for i in range(10):
        print(("STEP", i))
        downloaded = downloader.get_finished()
        for url in downloaded:
            print(('FINISHED:', url))
        if i > 1 and first:
            downloader.send_request(URLRequest('http://www.python.org/images/python-logo.gif'))
            downloader.send_request(URLRequest('http://www.python.org/'))
            downloader.send_request(URLRequest('http://www.doughellmann.com/PyMOTW/contents.html'))
            downloader.send_request(URLRequest('http://playermissile.com'))
            first = False
        time.sleep(1)

    downloader = None



import fleep

import logging
log = logging.getLogger(__name__)


def identify_loader(file_guess):
    info = fleep.get(file_guess.sample_data)
    if info.mime:
        if isinstance(info.mime, list):
            log.debug(f"fleep loader: identified multiple mime types {info.mime}; skipping")
            return None
        else:
            log.debug(f"fleep loader: identified {info.mime}")
            return dict(mime=info.mime)
    log.debug(f"fleep loader: unidentified")
    return None



import os
import sys
import json

import wx

from . import action
from . import errors
from . import clipboard
from .preferences import find_editor_preferences
from .utils import jsonutil
from .utils.command import DisplayFlags
from .utils.pyutil import get_plugins
from .menubar import MenuDescription
from .filesystem import fsopen as open

import logging
log = logging.getLogger(__name__)
clipboard_log = logging.getLogger("sawx.clipboard")
event_log = logging.getLogger("event")


def get_editors():
    return get_plugins('sawx.editors', SawxEditor)

def find_editor_class_for_document(document):
    """Find the "best" editor for a given MIME type string.

    First attempts all editors with exact matches for the MIME string,
    and if no exact matches are found, returns through the list to find
    one that can edit that class of MIME.
    """
    all_editors = get_editors()
    log.debug(f"find_editor_class_for_file: known editors: {all_editors}")
    matching_editors = []
    for editor in all_editors:
        try:
            state = editor.can_edit_document_exact(document)
        except Exception as e:
            log.debug(f"failure checking editor {editor}: {e}")
        else:
            if state:
                matching_editors.append(editor)
    log.debug(f"find_editor_class_for_file: exact matches: {matching_editors}")

    # # restore last session files from all possible editors before choosing best
    # # editor
    # for editor in matching_editors:
    #     editor.load_last_session(document)
    # for editor in matching_editors:
    #     if editor.can_restore_last_session(document):
    #         return editor
    if matching_editors:
        return matching_editors[0]

    # Try generic matches if all else fails
    for editor in all_editors:
        if editor.can_edit_document_generic(document):
            return editor
    raise errors.UnsupportedFileType(f"No editor available for {document}")

def find_editor_class_by_id(editor_id):
    """Find the editor class given its class name

    Returns the SawxEditor subclass whose `name` class attribute matches
    the given string.
    """
    editors = get_editors()
    log.debug(f"finding editors using {editors}")
    for editor in editors:
        if editor.editor_id == editor_id:
            return editor
        if editor_id in editor.compatibility_names:
            return editor
    raise errors.EditorNotFound(f"No editor named {editor_id}")


class SawxEditor:
    editor_id = "sawx_base_editor"
    ui_name = "Sawx Framework Base Editor"

    # list of alternate names for this editor, for compatibility with task_ids
    # from Sawx 1.0
    compatibility_names = []

    menubar_desc = [
        ["File",
            ["New",
                "new_blank_file",
                None,
                "new_file_from_template",
            ],
            "new_window",
            None,
            "open_file",
            ["Open Recent",
                "open_recent",
            ],
            None,
            "save_file",
            "save_as",
            None,
            "quit",
        ],
        ["Edit",
            "undo",
            "redo",
            None,
            "copy",
            "cut",
            "paste",
            None,
            "prefs",
        ],
        ["Help",
            "about",
            None,
            ["Debug",
                "show_debug_log",
                "widget_inspector",
                "raise_exception",
                "test_progress",
            ],
        ],
    ]

    keybinding_desc = {
        "new_file": "Ctrl+N",
        "open_file": "Ctrl+O",
        "save_file" : "Ctrl+S",
        "save_as" : "Shift+Ctrl+S",
        "cut": "Ctrl+X",
        "copy": "Ctrl+C",
        "paste": "Ctrl+V",
    }

    toolbar_desc = [
        "open_file", "save_file", None, "undo", "redo", None, "copy", "cut", "paste"
    ]

    statusbar_desc = [
        ["main", -1],
        ["debug", -1],
    ]

    module_search_order = ["sawx.actions"]

    session_save_file_header = ""

    session_save_file_extension = ""

    tool_bitmap_size = 0  # default size for platform

    preferences_module = "sawx.preferences"

    # each editor class uses a single preferences object for all instances, but
    # can't use normal class inheritance because subclasses will find an
    # already existing preference object if one in its superclass hierarchy
    # exists. So, we keep track of preference objects by class
    _preferences = {}

    # if an editor is marked as transient, it will be replaced if it's the
    # active frame when a new frame is added.
    transient = False

    # flag to indicate that the editor needs its own frame, without a tabbed
    # notebook and thus not shared with any other editor
    needs_tabness_frame = False

    #### class methods

    @classmethod
    def get_preferences(cls):
        if cls not in cls._preferences:
            prefs = find_editor_preferences(cls.preferences_module)
            prefs.restore_user_settings()
            log.debug(f"get_preferences: creating preferences: {prefs}")
            cls._preferences[cls] = prefs
        return cls._preferences[cls]

    #### dunder methods

    def __init__(self, document, action_factory_lookup=None):
        self.frame = None
        if action_factory_lookup is None:
            action_factory_lookup = {}
        self.action_factory_lookup = action_factory_lookup
        self.document = document
        self.last_loaded_uri = document.uri
        self.last_saved_uri = None
        self.show_toolbar = True

    def __str__(self):
        return f"{self.__class__.__name__}: document={self.document}, prefs={self.get_preferences()}"

    #### properties

    @property
    def is_dirty(self):
        return self.document.is_dirty

    @property
    def can_cut(self):
        return self.can_copy

    @property
    def can_copy(self):
        return False

    @property
    def can_paste(self):
        return self.frame.active_editor_can_paste

    @property
    def best_file_save_dir(self):
        attempts = []
        if self.last_saved_uri:
            # try most recent first
            attempts.append(self.last_saved_uri)
        if self.last_loaded_uri:
            # try directory of last loaded file next
            attempts.append(self.last_loaded_uri)
        if self.document is not None and self.document.uri:
            # path of current file is the final try
            attempts.append(self.document.uri)

        log.debug(("attempts for best_file_save_dir: %s" % str(attempts)))
        dirpath = ""
        for uri in attempts:
            uri_dir = os.path.dirname(uri)
            if os.path.exists(uri_dir):
                dirpath = uri_dir
                break

        return dirpath

    @property
    def title(self):
        if self.document is not None:
            uri = self.document.uri
        else:
            uri = self.last_saved_uri or self.last_loaded_uri
        if uri:
            if self.is_dirty:
                uri = "\u2605" + uri
            return uri
        return self.ui_name

    @property
    def tab_name(self):
        name = self.ui_name
        if self.document is not None:
            name = self.document.name
        if self.is_dirty:
            name = "\u2605" + name
        return name

    @property
    def is_transient(self):
        return self.transient or self.__class__ == SawxEditor

    @property
    def editor_summary(self):
        lines = []
        lines.append(f"name: {self.ui_name} ({self.editor_id})")
        lines.append(f"document: {self.document}" % self.document)
        lines.append(f"flags: is_dirty: {self.is_dirty}\n  can_cut: {self.can_cut}\n  can_copy: {self.can_copy}\n  can_paste: {self.can_paste}" )
        lines.append(f"last saved: {self.last_saved_uri}")
        lines.append(f"last loaded: {self.last_loaded_uri}")
        return "\n".join(lines)

    @property
    def preferences(self):
        return self.get_preferences()

    #### document destruction

    def prepare_destroy(self):
        """Release any resources held by the editor, but don't delete the main
        control as that will be deleted by the tabbed notebook control.
        """
        self.document.halt_background_processing()
        self.document.prepare_destroy()

    #### document creation

    def create_control(self, parent):
        return wx.StaticText(parent, -1, "Base class for Sawx editors")

    def create_layout(self):
        """Called after the control has been created to allow for initial setup
        of the user interface
        """
        pass

    def show(self, args=None):
        """populate the control with the editor's document.

        Can't do this at editor construction time because the control may not
        exist.
        """
        pass

    def create_event_bindings(self):
        pass

    def idle_when_active(self):
        """Called during idle events when this editor is the one active editor
        in the entire application, i.e. the active tab in the active frame.
        """
        pass

    #### file load/save

    def can_load_file(self, file_metadata):
        """Override in subclass if possible to load files into the document.
        """
        return False

    def load_file(self, file_metadata):
        """Override in subclass to actually load file into the document.
        """
        pass

    def load_success(self, path=None):
        if path is None:
            path = self.document.uri
        self.last_loaded_uri = path
        self.update_recent_path(path)
        self.frame.status_message(f"loaded {path}", True)

    def save_to_uri(self, uri=None, save_session=True):
        self.document.save(uri)
        if save_session:
            self.save_session()
        self.save_success()

    def save_as_image(self, uri, raw_data=None):
        """ Saves the contents of the editor in a maproom project file
        """
        valid = {
            '.png': wx.BITMAP_TYPE_PNG,
            '.tif': wx.BITMAP_TYPE_TIFF,
            '.tiff': wx.BITMAP_TYPE_TIFF,
            '.jpg': wx.BITMAP_TYPE_JPEG,
            '.jpeg': wx.BITMAP_TYPE_JPEG,
            }
        _, ext = os.path.splitext(uri)
        if ext not in valid:
            path += ".png"
            t = wx.BITMAP_TYPE_PNG
        else:
            t = valid[ext]

        if raw_data is None:
            raw_data = self.get_numpy_image()
        if raw_data is not None:
            h, w, depth = raw_data.shape
            image = wx.Image(w, h, raw_data)
            image.SaveFile(uri, t)

    def get_numpy_image_before_prompt(self):
        pass

    def get_numpy_image(self):
        raise NotImplementedError

    def save_session(self):
        s = {}
        self.serialize_session(s)
        self.document.save_session(self.editor_id, s)

    def save_success(self, path=None):
        if path is None:
            path = self.document.uri
        self.last_saved_uri = path
        self.update_recent_path(path)
        self.frame.status_message(f"saved {path}", True)

    def update_recent_path(self, path):
        from sawx.actions import open_recent
        open_recent.open_recent.append(path)
        self.frame.status_message(f"saved {path}", True)

    #### search

    @property
    def search_start(self):
        """Return index at which the next search will start.
        """
        return 0

    #### popup menu utilities

    def show_popup(self, popup_menu_desc, popup_data=None):
        valid_id_map = {}
        menu = MenuDescription("popup", popup_menu_desc, self, valid_id_map, popup_data=popup_data)
        menu.sync_with_editor(valid_id_map)
        action_id = self.control.GetPopupMenuSelectionFromUser(menu.menu)
        if action_id == wx.ID_NONE:
            log.debug("show_popup: cancelled")
        else:
            action_key, action = valid_id_map[action_id]
            log.debug(f"show_popup: id={action_id}, action_key='{action_key}' action={action}")
            try:
                wx.CallAfter(action.perform, action_key)
            except AttributeError:
                log.warning(f"no perform method for {action}")

    #### clipboard operations

    supported_clipboard_handlers = []

    @property
    def supported_clipboard_data(self):
        return [c[0] for c in self.supported_clipboard_handlers]

    def get_clipboard_handler(self, data_obj):
        for d, handler_name in self.supported_clipboard_handlers:
            if d == data_obj:
                try:
                    handler = getattr(self.__class__, handler_name)
                except AttributeError:
                    handler = getattr(clipboard, handler_name)
                return handler

    def copy_selection_to_clipboard(self):
        focused = wx.Window.FindFocus()
        clipboard_log.debug(f"focused control: {focused}")
        try:
            data_objs = self.calc_clipboard_data_from(focused)
        except ValueError:
            log.warning("No selection")
        else:
            clipboard_log.debug(f"created data objs: {data_objs}")
            clipboard.set_clipboard_data(data_objs)

    def calc_clipboard_data_from(self, focused):
        return clipboard.calc_data_objects_from_control(focused)

    def delete_selection(self):
        focused = wx.Window.FindFocus()
        log.debug(f"deleting selection from {focused}")
        self.delete_selection_from(focused)

    def delete_selection_from(self, focused):
        pass

    def paste_clipboard(self):
        focused = wx.Window.FindFocus()
        clipboard_log.debug(f"focused control: {focused}")
        data_obj = clipboard.get_clipboard_data(self.supported_clipboard_data)
        if data_obj:
            handler = self.get_clipboard_handler(data_obj)
            if handler:
                handler(self, data_obj, focused)
            else:
                clipboard_log.error("No clipboard handler found for data {data_obj}")

    #### command processing

    def calc_status_flags(self):
        return DisplayFlags()

    def undo(self):
        undo = self.document.undo_stack.undo(self)
        self.process_flags(undo.flags)
        self.frame.sync_active_tab()

    def redo(self):
        undo = self.document.undo_stack.redo(self)
        self.process_flags(undo.flags)
        self.frame.sync_active_tab()

    def end_batch(self):
        self.document.undo_stack.end_batch()

    def process_command(self, command, batch=None):
        """Process a single command and immediately update the UI to reflect
        the results of the command.
        """
        event_log.debug(f"\nprocess_command: starting command {command}")
        f = self.calc_status_flags()
        undo = self.process_batch_command(command, f, batch)
        if undo.flags.success:
            self.process_flags(f)
        event_log.debug(f"process_command: finished command {command}\n")
        return undo

    def process_batch_command(self, command, f, batch=None):
        """Process a single command but don't update the UI immediately.
        Instead, update the batch flags to reflect the changes needed to
        the UI.
        
        """
        undo = self.document.undo_stack.perform(command, self, batch)
        f.add_flags(undo.flags, command)
        return undo

    def process_flags(self, flags):
        """Perform the UI updates described by the DisplayFlags instance
        
        """
        log.debug("processing flags: %s" % str(flags))
        d = self.document
        visible_range = False

        if flags.message:
            self.frame.status_message(flags.message)

        if flags.rebuild_ui:
            log.debug(f"process_flags: rebuild_ui")
            d.recalc_event(flags=flags)
        if flags.refresh_needed:
            log.debug(f"process_flags: refresh_needed")
            d.recalc_event(flags=flags)


    #### session

    def get_editor_specific_metadata(self, keyword):
        """Get the metadata for the keyword that is specific to this editor
        class.

        E.g. if there is a layout stored in the metadata, it will be in
        metadata[self.editor_id]['layout']:

        {
            "omnivore.byte_edit": {
                "layout": {
                    "sidebars": [ ... ],
                    "tile_manager": { .... },
                }
        }
        """
        try:
            layout_dict = metadata[self.editor_id][keyword]
        except KeyError:
            layout_dict = {}
        return layout_dict

    def serialize_session(self, s):
        pass

    #### file identification routines

    @classmethod
    def can_edit_document_exact(cls, document):
        return False

    @classmethod
    def can_edit_document_generic(cls, document):
        return False

    @classmethod
    def can_edit_document(cls, document):
        return cls.can_edit_document_exact(document) or cls.can_edit_document_generic(document)

    def calc_usable_action(self, action_key):
        try:
            action_factory = self.action_factory_lookup[action_key]
        except KeyError:
            if "{" in action_key and action_key.endswith("}"):
                action_key_lookup, _ = action_key[:-1].split("{", 1)
            else:
                action_key_lookup = action_key
            action_factory = action.find_action_factory(self.module_search_order, action_key_lookup)
            self.action_factory_lookup[action_key] = action_factory
        return action_factory(self, action_key)
